#!/usr/bin/perl
#
#use diagnostics;
#use CGI qw/:standard :html4/;
#use CGI qw/:standard :html4 :form :cgi :netscape/;
use CGI qw(:all);

#################################################################
# glinks.pl
# GRASSLinks                           	Version 3.2beta
# Additions by Len Coop                 Last Modified 9/01/2002
# GRASSLinks 3.2beta Copyright (c) 2002 
# Oregon State University/Integrated Plant Protection Center
# See license and disclaimer for GRASSLinks 3.1 below
#
# Changelog version 3.2beta - L. Coop IPPC Oregon State University
# See glinks/CHANGELOG for complete list
#
# Changes planned/needed: See glinks/TODO for complete list
#
# Brief Summary: GRASSLinks is a perl web interface to GRASS GIS allowing
#  display, zoom, pan, query and advanced analysis of spatial data 
#  The goal of GRASSLinks is to develop a simple to use, powerful
#  web interface to GIS and mapping. Compare this program to 
#  mapserver (no analysis capabilties) and Grules (C not perl
#  also mainly for display of GRASS data).
################################################################
# GRASSLinks                           	Version 3.1.1
# Created by Susan M. Huse, 1994        Last Modified 05/06/98 
# tested / runs under perl 4.0, 5.002
#################################################################
#################################################################
# Grasslinks 3.1 Copyright (c) 1998 
# The Regents of the University of California (Regents).
# Created by Patricia Frontiera, Susan Huse, James Ganong, and Kenn Gardels,
# Center for Environmental Design Research, University of California, Berkeley.
# GRASSLinks 3.1 is based on a work created by 
# Susan Huse at the University of California, Berkeley.
# License:
# Redistribution and use of GRASSLinks 3.1 in source and binary forms,
# with or without modification, are permitted provided that the
# following conditions are met: 1.Redistributions of source code (with
# or without modification) must retain the above copyright notice, this
# list of conditions and the following disclaimer. 2.Redistributions in
# binary form (with or without modification) must reproduce the above
# copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with
# the distribution.
#
# IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
# INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
# LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
# DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF
# ANY, PROVIDED HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO
# OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
# ANY, PROVIDED HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO
# OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
# MODIFICATIONS.
#
#################################################################
#################################################################
# ERROR LOGGING note:
# GL errors written to STDERR which is the web server error
# log.  You could set up a temp file and write debug messages
# to it if needed.
#
# GIS errors are sent to the specified email address. If no map
# appears, check the email. 
#################################################################

require "flush.pl";    #flush buffer/forces print statements to STDERR & STDOUT
do "./glconf.pl";  #now nearly all user config settings located here



open( SAVEOUT, ">&STDOUT" );
select(STDOUT);
$| = 1;

&set_glvars;           # set GL web variables
&parse_vars;           # read input variables from the client form
&read_gldata;          # read in GL data file info
&set_gloptions;

# GRASSLINKS ACCESS LOG FILE
open( ACCESSLOG, ">>$GL_ACCESS_LOG" ) if defined $GL_ACCESS_LOG;

&start_GRASS;          # start grass preamble
    #print "Content-type: text/html\n\n did start\n";

# process and/or exit
#my $pid = $$;
if ( $pid = fork ) {
    wait();
    close(STDOUT);
    open( STDOUT, ">&SAVEOUT" );
    $SIG{'ALRM'} = 'IGNORE';
    &stop_GRASS;
    close(ACCESSLOG) if defined $GL_ACCESS_LOG;

    if ( -e $error_tmp ) {
        system("cat $error_tmp");
    }
    else {
        system("cat $html_tmp");
    }
}
else {
    setsid;
    $SIG{'ALRM'} = 'roundup';
    alarm($TIME_OUT);

    open( STDOUT, ">$html_tmp" );

    # run the appropriate subroutine
    eval "&$value{ANALYSIS}";

}
close(SAVEOUT);
exit 0;

#################################################################
# End of main GRASSLinks routine				#
# everything from here down is subroutines			#
#################################################################

#################################################################
# ERROR CHECKING SUBROUTINES & BASIC UTILITIES			#	
#################################################################

#-------------------------------------------------------------------------
# NAME: roundup	
# 	This routine should exit gracefully in the event of bad
# 	process or timeout triggered.
#	Kills child processes that might linger (I hope).
#-------------------------------------------------------------------------
sub roundup {

    $SIG{'ALRM'} = 'IGNORE';

    if ( defined $GL_ACCESS_LOG ) {
        print ACCESSLOG
"GL ROUNDUP: Timeout alarm triggered in $value{ANALYSIS}, $TIME_OUT seconds\n";
    }
    else {
        print STDERR
"GL ROUNDUP: Timeout alarm triggered in $value{ANALYSIS}, $TIME_OUT seconds\n";
    }

    &error(
        "TIMEOUT ERROR: command ran for $TIME_OUT seconds and will be killed");
    kill HUP => -$$;
}

#-------------------------------------------------------------------------
# NAME: system_errors_logged	
#	System call wrapper. STDERR sent to web server log file.
#-------------------------------------------------------------------------
sub system_errors_logged {
    system("$_[0]");
}

#-------------------------------------------------------------------------
# NAME: system_silent
#	System call wrapper; squelches any noise.
#-------------------------------------------------------------------------
sub system_silent {
    system("$_[0] > /dev/null 2>&1  ");
}

#-------------------------------------------------------------------------
# NAME: error
# 	Creates error / warning html to predefined $error_tmp file
#	which is returned to user when program exits. 
# ARGS: $_[0] = Error message string.
#-------------------------------------------------------------------------
sub error {
    my $errorMsg = $_[0];
    open( OUTPUT, ">$error_tmp" )
      || print STDERR "GL can't open $error_tmp to create error html file\n";
    print OUTPUT "Content-type: text/html\n\n";
    print OUTPUT << "ENDPRINT";
        <HTML>
        <HEAD>
        <TITLE>GRASSLinks Error</TITLE>
        </HEAD>
        <BODY>
	<h2>GRASSLINKS Error:</h2>
	<hr>
	<br>
	<h3>$errorMsg </h3>
	<br><hr><br>
	<b>Back up your Web browser, review your selection(s) and try again!</b>
	</BODY>
	</HTML>
ENDPRINT
    close(OUTPUT);
    &flush(STDERR);
}

#-------------------------------------------------------------------------
# NAME: print_file	
#	Print out contents of file into html file.
# ARGS: $_[0] = The name of the file to print out.
#-------------------------------------------------------------------------
sub print_file {
    my $thefile = $_[0];

    if ( -e $thefile ) {
        open( THE_FILE, "$thefile" )
          || print STDERR
          "In GL print_file subroutine and can't open file .$thefile.\n";
        while (<THE_FILE>) {
            print $_;
        }
        close(THE_FILE);
        &flush(STDERR);
    }
    else {
        print STDERR "In GL print_file, can't find file .$thefile.\n";
    }
}

#################################################################
# Variable Assignment Subroutines				#
#################################################################

#-------------------------------------------------------------------------
# NAME:	parse_vars
#	This subroutine reads in variables in either POST or GET
# 	method, checks input values, and then processes selected
#	analysis option.
# 	SECURITY ESSENTIAL - make no changes unless you are sure
# 	how they may effect security!
# REQUIRES: A valid http GET or POST input stream that contains
# 	at least a valid value for $value{PROJECT} and $value{ANALYSIS}
#	(validity checked in this subroutine).
# RETURNS: Input stream read into Hash of variables $value{}
#	and into Hash of Arrays $Values{}.
#-------------------------------------------------------------------------
sub parse_vars {

    my ( $in, @input, $input, $name, $val, $goodval );

    # read in input stream	
    # eventually replace this with cgi-lib.pl
    if ( $ENV{REQUEST_METHOD} eq 'POST' ) {
        read( STDIN, $input, $ENV{'CONTENT_LENGTH'} );
        $input =~ s/%(..)/pack("C",hex($1))/eg;
        @input = split ( /&/, $input );

        #UNCOMMENT FOR DEBUGGING
        #foreach $i (@input) { print STDERR "POST input: $i \n";}
    }
    elsif ( $ENV{REQUEST_METHOD} eq 'GET' ) {
        $in = $ENV{'QUERY_STRING'};
        $in =~ s/%(..)/pack("C",hex($1))/eg;
        @input = split ( /\+/, $in );

        #UNCOMMENT FOR DEBUGGING
        #foreach $i (@input) { print STDERR "GET input: $i \n";}
    }
    else {
        &error("Invalid input, exiting!");
        return;
    }

    foreach $i (@input) {    #create assoc. array to hold forms variables
        ( $name, $val ) = split ( /=/, $i );
        if ( length($val) > 99 ) {
            &error(
                "Form input variable $name longer than 99 characters; exiting."
            );
            return;
        }

        #remove all potentially unsafe characters
        $val =~ tr/ /_/;                   #no spaces
        $val =~ tr/+/_/;                   #no pluses
        $val =~ s/[^a-zA-Z0-9_.,@-]//g;    #anything but these strip
             #reassign val to goodval to be extra safe
        $goodval = $val;

        if ($goodval) {
            $value{$name} = $goodval;

            # Values{} holds multiple values- HashOfArrays 
            push ( @{ $Values{$name} }, $goodval );
        }

    }    #end of foreach to process name-value associative array

    # For security reasons, make sure the value of $value{ANALYSIS}, which
    # launches the main GL routine, is a valid one as set in the
    # subroutine set_glvars

    unless ( $GL_ANALYSES{ $value{ANALYSIS} } ) {
        &error("Analysis value \"$value{ANALYSIS}\" not recognized.");
        return;
    }

    #  Same security chec for valid project codes
    unless ( $GL_PROJECTS{ $value{PROJECT} } ) {
        &error("Project value \"$value{PROJECT}\" not recognized.");
        return;
    }

    &flush(STDERR);

}    #end of sub parse_vars


#-------------------------------------------------------------------------
# NEXT 2 subroutines now in glconf.pl to help simplify user config
#
#-------------------------------------------------------------------------
# NAME: set_glvars
# 	Sets variables needed to run GL on this web server set.  
# REQUIRES: Must be configured for local site.
#-------------------------------------------------------------------------
#-------------------------------------------------------------------------
# NAME: set_GRASSvars
# 	Sets variables specific to GRASS as it runs under GL.
# REQUIRES: Must be configured for local site.
#-------------------------------------------------------------------------


#-------------------------------------------------------------------------
# NAME: read_gldata 
#	Reads in from file grass data files available in GL
# 	depending on a project variable.
# REQUIRES: The project variable $value{PROJECT} must be set
#	and a file must exist that lists available GRASS raster,
#	vector, sites, and region files, as well as specifics
# 	about the current GL project. 
# NOTE: This subroutine expects each GRASS file listed in the $GLDATA_LIST
#	file to be of the following format: 1) fields are delimited by the
#	"|" character; 2) field 0 always contains the file type: raster,
#	vector, sites, region, and project - which is not a grass file type! 
#	3) each file has info listed about it in a determined order as 
#	indicated in this subroutine in the SWITCH block
#-------------------------------------------------------------------------
sub read_gldata {

    my ( @thefile, @data, $line );

    open( LIST, "./GL_list" )
      || print STDERR
      "GL unable to open $GLDATA_LIST in subroutine read_gldata\n";

    @thefile = <LIST>;
    close(LIST);

    foreach $line (@thefile) {
        chop $line;

        # load into temp array @data
        @data = split ( /\|/, $line );
        for ( $i = 0 ; $i < 4 ; $i++ ) {
            $data[$i] = "none" unless defined $data[$i];
        }

        # only grab files for current project
        next unless $data[1] =~ /$value{PROJECT}/ || $data[1] =~ /\!/;

        SWITCH: {
            if ( $line =~ /^raster\|/ ) {
                $rec = {};
                $rec->{projects}  = $data[1];
                $rec->{filename}  = $data[2];
                $rec->{descShort} = $data[3];

                # the following 3 fields aren't used in
                # this script but they could be. They
                # are used in helper GL scripts
                #$rec->{mdata} = $data[4];
                #$rec->{gif} = $data[5];
                #$rec->{descLong} = $data[6];
                #whole thing in HoH
                $RASTERS{ $rec->{filename} } = $rec;

                #whole thing in HoA
                push @{ $GLfiles{raster} }, [@data];
                last SWITCH;
            }
            if ( $line =~ /^vector\|/ ) {
                $rec = {};
                $rec->{projects}   = $data[1];
                $rec->{filename}   = $data[2];
                $rec->{descShort}  = $data[3];
                $rec->{vectorform} = $data[6];

                #whole thing into HoH
                $VECTORS{ $rec->{filename} } = $rec;

                #whole thing in HoA
                push @{ $GLfiles{vector} }, [@data];
                last SWITCH;
            }
            if ( $line =~ /^sites\|/ ) {
                $rec = {};
                $rec->{projects}  = $data[1];
                $rec->{filename}  = $data[2];
                $rec->{descShort} = $data[3];
                $rec->{siteform}  = $data[6];

                #whole thing
                $SITES{ $rec->{filename} } = $rec;

                #whole thing in HoA
                push @{ $GLfiles{sites} }, [@data];
                last SWITCH;
            }
            if ( $line =~ /^region\|/ ) {
                $rec = {};
                $rec->{filename}  = $data[2];
                $rec->{descShort} = $data[3];

                #whole thing
                $REGIONS{ $rec->{filename} } = $rec;

                #whole thing in HoA
                push @{ $GLfiles{region} }, [@data];
                last SWITCH;
            }
            if ( $line =~ /^project\|/ ) {

                # only one project per GL session
                $project{title}    = $data[2];
                $project{gif}      = $data[3];
                $project{location} = $data[4];
                $project{region}   = $data[5];
                last SWITCH;
            }
        }    # end SWITCH BLOCK

    }    # end foreach

}    # end of  subroutine

#-------------------------------------------------------------------------
# NAME: set_gloptions 
#	Indicates list of values for specific GL options.
# 	If you want to indicate a default select value, then
#	you must indicate by placing a "!!" in field [1].
#-------------------------------------------------------------------------
sub set_gloptions {

    my ( @color, @ssize, @gifsize, @legend, @repUnits, @bufUnits, @resValues, @opacity );

    @color = (
        [ "color", "!!", "black",   "black" ],
        [ "color", "!",  "white",   "white" ],
        [ "color", "!",  "red",     "red" ],
        [ "color", "!",  "orange",  "orange" ],
        [ "color", "!",  "yellow",  "yellow" ],
        [ "color", "!",  "green",   "green" ],
        [ "color", "!",  "blue",    "blue" ],
        [ "color", "!",  "indigo",  "indigo" ],
        [ "color", "!",  "brown",   "brown" ],
        [ "color", "!",  "magenta", "magenta" ],
        [ "color", "!",  "aqua",    "aqua" ],
        [ "color", "!",  "grey",    "grey" ],
    );
    $GLfiles{color} = [@color];

    @ssize = (
        [ "ssize", "!",  "invis_point", "invisible" ],    # all _point icons
        [ "ssize", "!",  "vsm_point",   "tiny icon" ],    # created with p.icons
        [ "ssize", "!",  "sm_point",    "v. small icon" ],
        [ "ssize", "!",  "med_point",   "small icon" ],
        [ "ssize", "!",  "lg_point",    "med. icon" ],
        [ "ssize", "!",  "vlg_point",   "large icon" ],
        [ "ssize", "!!", "1",           "small" ],        # these are built-in
        [ "ssize", "!",  "2",           "medium" ],
        [ "ssize", "!",  "3",           "large" ],
    );
    $GLfiles{ssize} = [@ssize];

    @gifsize = (
        [
            "gifsize", "!",
            "400",     "small (400X400 -- est. processing time 4 seconds)"
        ],
        [
            "gifsize", "!!",
            "600",     "medium (600X600 -- est. processing time 5 seconds)"
        ],
        [
            "gifsize", "!",
            "750",     "med. lg. (750X750 -- est. processing time 6 seconds)"
        ],
        [
            "gifsize", "!",
            "1000",    "large (1000X1000 -- est. processing time 9 seconds)"
        ],
        [
            "gifsize",
            "!",
            "250",
            "thumbnail (250X250 -- est. processing time 3 seconds)"
        ],
        [
            "gifsize",
            "!",
            "150",
            "sm. thumbnail (150X150 -- est. processing time 3 seconds)"
        ]
    );
    $GLfiles{gifsize} = [@gifsize];

    @legend = (
        [
            "legend",
            "!",
            "yes",
            "Create a Legend for the Map (increases processing time"
        ],
        [ "legend", "!!", "no", "Do Not Create a Legend for the Map" ],
    );
    $GLfiles{legend} = [@legend];

    @repUnits = (
        [ "report_units", "!",  "mi", "square miles" ],
        [ "report_units", "!",  "me", "square meters" ],
        [ "report_units", "!",  "a",  "acres" ],
        [ "report_units", "!",  "h",  "hectares" ],
        [ "report_units", "!!", "k",  "square kilometers" ],
    );
    $GLfiles{repUnits} = [@repUnits];

    @bufUnits = (
        [ "buff_units", "!!", "kilometers", "kilometers" ],
        [ "buff_units", "!",  "meters",     "meters" ],
        [ "buff_units", "!",  "feet",       "feet" ],
        [ "buff_units", "!",  "miles",      "miles" ],
    );
    $GLfiles{bufUnits} = [@bufUnits];

    @resValues = (
        [ "resolution", "!",  "5", "coarse" ],
        [ "resolution", "!!", "3", "medium" ],
        [ "resolution", "!",  "1", "fine" ],
    );
    $GLfiles{resValues} = [@resValues];

    @opacity = (
        [ "opacity", "!",  "1.0", "zero" ],
        [ "opacity", "!",  "0.75", "minimal" ],
        [ "opacity", "!",  "0.6", "low" ],
        [ "opacity", "!!", "0.5", "medium" ],
        [ "opacity", "!",  "0.4", "med. high" ],
        [ "opacity", "!",  "0.25", "v. high" ],
        [ "opacity", "!",  "0.0", "total" ],
        [ "opacity", "!",  "2.0", "variable" ],
    );
    $GLfiles{opacity} = [@opacity];

}    # end of set_gloptions  subroutine

#################################################################
# GRASS base subroutines					#
#################################################################

#-------------------------------------------------------------------------
# NAME: lock_Gmapset 
# 	Steps through all grasslinks mapsets until find an available one.
#	A mapset is available if there exists in the mapset the file
#	UNLOCK.  The mapset is unavailable if the file is LOCK.
# REQUIRES: This subroutine assumes that foreach number $N_GL_USERS
#	there exists a grass mapset directory in the name form:
#	$ENV{GISDBASE}/$ENV{LOCATION_NAME}/grasslinks$locnumber
#	and that only $N_GL_USERS can run grasslinks simultaneously
#	the UNLOCK file must be manually setup in mapset when GL first setup
#	and the LOCK files may need to be renamed UNLOCK if GL bombs 
# RETURNS: A mapset locked for user. Must be unlocked by the
#	subroutine stop_GRASS when user exits GL.
#-------------------------------------------------------------------------
sub lock_Gmapset {

    #$ENV{LOCATION_NAME} = "spearfish";
    #$LOCATION_NAME = "spearfish";
    my $location  = "$ENV{GISDBASE}/$ENV{LOCATION_NAME}/grasslinks";
    my $locnumber = 1;
    if ( defined $GL_ACCESS_LOG ) { print ACCESSLOG "loc= $location\n"; }

    while ( $locnumber <= $N_GL_USERS ) {
        if ( defined $GL_ACCESS_LOG ) {
            print ACCESSLOG "locnum= $locnumber\n";
        }
        if (
            rename( "$location$locnumber/UNLOCK", "$location$locnumber/LOCK" ) )
        {

            #the mapset is available and locked
            $ENV{MAPSET}   = "grasslinks$locnumber";
            $ENV{LOCATION} = "$ENV{GISDBASE}/$ENV{LOCATION_NAME}/$ENV{MAPSET}";

            #$ENV{GISRC} = "$ENV{LOCATION}/.grassrc5";
            $ENV{GISRC} = "$ENV{LOCATION}/.grassrc5";
            $ENV{MAPLP} = "$ENV{LOCATION}/pmap.$ENV{MAPSET}";
            $MAPSET = $ENV{MAPSET};
            if ( defined $GL_ACCESS_LOG ) {
                print ACCESSLOG "MAPSET= $MAPSET\n";
            }
            $LOCATION = "$ENV{GISDBASE}/$ENV{LOCATION_NAME}/$MAPSET";
            my $GISRC = "";
            $GISRC = "$ENV{LOCATION}/.grassrc5";
            print ACCESSLOG "GISRC= $GISRC\n";
            my $MAPLP = "";
            $MAPLP = $ENV{MAPLP};

            # Update Access Log
            if ( defined $GL_ACCESS_LOG ) {
                $theTime = `date`;
                chomp($theTime);
                print ACCESSLOG
"GL STARTED ($value{ANALYSIS}) at $theTime by: $ENV{REMOTE_ADDR} in mapset $MAPSET\n";

#print STDERR "GL STARTED ($value{ANALYSIS}) at .$theTime. by: .$ENV{REMOTE_USER}.$ENV{REMOTE_ADDR}.$ENV{REMOTE_HOST}.$ENV{REMOTE_IDENT} in mapset .$MAPSET.\n"; 
                &flush(ACCESSLOG);
            }
            last;

        }
        else {

            #that mapset is not available, move to the next
            $locnumber = $locnumber + 1;
            if ( $locnumber > $N_GL_USERS ) {
                &error(
"All grasslinks mapsets are busy.  Try again later, or press the back button and retry now.\n"
                );
                return;
            }
        }
    }    #end of while loop

    &flush(STDERR);
}

#-------------------------------------------------------------------------
# NAME: start_GRASS 
# 	Sets necessary GRASS variables, locks a grass mapset, and
#	removes any set MASK.	
#-------------------------------------------------------------------------
sub start_GRASS {

    &set_GRASSvars;    #set grass specific variables
    print ACCESSLOG "in start_GRASS-set_GRASSvars\n";
    &lock_Gmapset;     #lock a GL mapset
    print ACCESSLOG "in start_GRASS-lock_Gmapset\n";

    #just in case a grass mask exists, remove it!
    #&system_silent("g.remove rast=MASK"); 
}

#-------------------------------------------------------------------------
# NAME: stop_GRASS 
#	Unlocks a grass mapset.
#-------------------------------------------------------------------------
sub stop_GRASS {

    rename( "$LOCATION/LOCK", "$LOCATION/UNLOCK" )
      || print STDERR "GL can't unlock mapset in $LOCATION/LOCK\n";
    if ( defined $GL_ACCESS_LOG ) {
        print ACCESSLOG
          "GL FINISHED ($value{ANALYSIS}): unlocked mapset $MAPSET\n";

#print STDERR "GL FINISHED ($value{ANALYSIS}): unlocked mapset $MAPSET\n";
    }

    &flush(STDERR);
}

#################################################################
# MAIN subroutines for each GRASSLinks analysis option		#
#################################################################

#-------------------------------------------------------------------------
# NAME: display1 
#	This subroutine creates select menus for display request form.
#	Items displayed in select lists are from the $GLDATA_LIST file
#	and are a function of the current $value{PROJECT} code.
#-------------------------------------------------------------------------
sub display1 {

    &begin_html( "Display Request Form", "FORM" );

    print << "ENDPRINT";
	<b>&#42;</b> To create a <b>map display</b>, select from
	the menus of <a href="$HTML_PATH/$GL_MATRIX" target="new">
	GRASS map layers</a> shown below.
	<br>
	<b>&#42;</b> Refer to the <a href="$GL_HELP#display" target="new">Help 
	document</a> if you need further information.
	<br>
	<b>&#42;</b> See bottom of this page for <a href="#Advanced">
	Advanced GRASSLinks Options</a>.
	<BR>
	<hr>
	<li><b>Select a single, raster map layer for display, zoom, pan, and query:
	</b><BR>
ENDPRINT

    &make_select( "RASTER", "4", "raster", "" );
    print '
	<b><li>Select an overlay raster map and opacity value:
	</b><BR>
	';

    &make_select( "RASTER_SUB", "4", "raster", "" );
    print ' <b>opacity: </b> ';



    &make_select( "SUB_OPACITY", "1", "opacity", "" );

    print '
	<BR><hr><b>
	<li>Select any number of vector layers to overlay:
	</b><br>
	';

    &make_select( "VECTOR", "4", "vector", "MULTIPLE" );

    print '
	<BR>
	Select colors for the first four vector layers 
	(any others selected will default to black):
	<br>
	';

    &make_select( "VCOLOR1", "1", "color", "" );
    &make_select( "VCOLOR2", "1", "color", "" );
    &make_select( "VCOLOR3", "1", "color", "" );
    &make_select( "VCOLOR4", "1", "color", "" );

    print '
	<hr><li><b>
	Select any number of site layers to display as 
	overlay points:
	</b><br>
	';

    &make_select( "SITES", "3", "sites", "MULTIPLE" );

    print '
	<BR>
	Select colors for up to the first three site layers 
	(any others selected will default to black):
	<br>
	';

    &make_select( "SCOLOR1", "1", "color" );
    &make_select( "SCOLOR2", "1", "color" );
    &make_select( "SCOLOR3", "1", "color" );

    print '
	<BR>
	Select site marker size for up to the first three site layers 
	(any others selected will default to medium size):
	<br>
	';

    &make_select( "SSIZE1", "1", "ssize" );
    &make_select( "SSIZE2", "1", "ssize" );
    &make_select( "SSIZE3", "1", "ssize" );

    print '
	<li><b>Option to display site values/labels:</b>
	<INPUT TYPE=RADIO NAME="SITELABELS" and VALUE="no" CHECKED> off 
	<INPUT TYPE=RADIO NAME="SITELABELS" and VALUE="yes"> on
	<br>
	';

    print '
	<hr>
	<li><b>Select a region to display:</b>
	<br>
	';

    &make_select( "REGION", "4", "region", "" );

    print '
	<BR><li>
	<b>and an image size for the map:
	</b><br>
	';

    &make_select( "GIFSIZE", "3", "gifsize", "" );

    print '
	<BR><li><b>
	and if you want a legend for the map:
	</b><br>
	<INPUT TYPE="radio" NAME="LEGEND" VALUE="no" CHECKED> none
	<INPUT TYPE="radio" NAME="LEGEND" VALUE="yes"> all 
	<INPUT TYPE="radio" NAME="LEGEND" VALUE="samp"> some - not all
	';

    #&make_select("LEGEND", "2", "legend", "");

    print '<i> Note: Legends are only created for raster maps.</i>
	<br><hr>
	';


    # new 3-11-02 allow for custom title to appear above map
    print '
        <BR><li>
        <b>Enter a title for your map (optional - up to 44 characters):
        </b><br>
        <INPUT TYPE="TEXT" NAME="MAP_TITLE" SIZE=40 MAXLENGTH=64>
        <br>
        ';

    &make_hidden( "ANALYSIS", "display2" );
    &make_hidden("PROJECT");
    &end_ourform( "Create the Display Map", "Reset" );
    &end_ourhtml;
}

#-------------------------------------------------------------------------
# NAME: display2 
#	Create gif image of map layers and options selected by the
#	user in subroutine display1 and display in an html file.
#-------------------------------------------------------------------------
sub display2 {

    my ( $mapDesc, $regionInfo );

    if ( ( !$value{RASTER} ) && ( !$value{VECTOR} ) && ( !$value{SITES} ) ) {
        &error("No Map Layers chosen for Display!\n");
        return;
    }

    # set defaults image size and region
    $value{GIFSIZE}    = $DEFAULT_GIFSIZE    unless defined $value{GIFSIZE};
    $value{REGION}     = $DEFAULT_REGION     unless defined $value{REGION};
    $value{SITELABELS} = $DEFAULT_SITELABELS unless defined $value{SITELABELS};

    # file to contain script needed to create gif of map layer
    $value{PMAP_SCRIPT} = "pmap_script.$$";

    # gif file of map layer, must be in an htdocs directory
    $value{GIF} = "grasslinks.$$.gif";
    $value{PNG} = "grasslinks.$$.png";
    $value{PGM} = "grasslinks.$$.pgm";
    $value{PNM} = "grasslinks.$$.pnm";

    # file of description of map gif
    $value{DISPLAY_DESC} = "gl.dispDesc.$$";

    # make the pmap file need to create a GRASS Map
    $mapDesc = &make_pmap;

    # Get some more description info
    $regionInfo = $REGIONS{ $value{REGION} }->{descShort};
    $mapDesc .= "\n<FONT SIZE=-1><li>Region: <b>$regionInfo</b></li></FONT>";

# dont show image size - who cares?
#	$mapDesc .= "\n<FONT SIZE=-1><li>Image Size: <b>$value{GIFSIZE} x $value{GIFSIZE}</b></li></FONT>";

    #save map display description text for re-use
    open( MAP_DESC, ">$TMP/$value{DISPLAY_DESC}" );
    print MAP_DESC "\n<FONT SIZE=-1 COLOR=GREEN><b>Items Mapped:</b></FONT>$mapDesc</ul>";

    close(MAP_DESC);

    # set the region & create the GIF 
    &set_region;
    &get_coords;
    &adjust_region;
    &create_gif;

    # return map display to browser 
    #&begin_html( "Map Display, Zoom & Query", "FORM" );
    &begin_html( "Map Query", "FORM" );
    &create_display_html("","");
    &end_ourform;
    &end_ourhtml;
}

#-------------------------------------------------------------------------
# NAME: display3
# 	Either query a raster and sites maps to return
# 	attribute & coordinate info for the user's selected mousepoint
# 	OR zoom-in, zoom-out, or pan and recreate gif image. 
#	All depends on value of $value{ZOOM}. 
# 	This whole subroutine depends on existence of temp files
# 	made by earlier subroutines. But a cronjob routinely cleans
# 	temp files more than one hour old.
#-------------------------------------------------------------------------
sub display3 {
# invoke any extra forms that getted triggered by vector and site queries
    do "$GL_FORMS";

    my (
        $newMsg,     $cmd,         @sites_file,  $sites_detail,
        @siteinfo,   $siteforms,   @vector_file, $vector_detail,
        @vector_all, $vector_line, @vectorinfo,  $vectorforms,
        @catin,      @catdata
    );

    # First test for existance of key temp files!
    my $fileOne   = "$TMP/$value{DISPLAY_DESC}";
    my $fileTwo   = "$TMP/$value{PMAP_SCRIPT}";
    #my $fileThree = "$TMP/$value{GIF}";
    # raster sub is optional so comment out
    # my $fileThree = "$TMP/$value{PMAP_SUB_SCRIPT}";
    my $fileFour  = "$TMP/$value{PNG}";

    if ( !( -r $fileOne && -r $fileTwo  && -r $fileFour ) ) {
        &error("Sorry, your GRASSLinks session has timed out!\n");
        return;
    }
    else {
        &system_silent("touch $fileOne $fileTwo $fileFour");
    }

    #make sure ZOOM has a value, set to 0 as default if not defined
    if ( !$value{ZOOM} ) { $value{ZOOM} = 0; }

    # NEW compute scale and legend size factors based on zoom value
    if ( $value{ZOOM} == 0 ) { $scalefactor = 0.0; }
    if ( $value{ZOOM} == 1 ) { $scalefactor = 0.0; }

    #if ($value{ZOOM} == 2) { $scalefactor = 0.001;}
    if ( $value{ZOOM} == 2 )  { $scalefactor = 0.5; }
    if ( $value{ZOOM} == 3 )  { $scalefactor = 0.066; }
    if ( $value{ZOOM} == 5 )  { $scalefactor = 0.01; }
    if ( $value{ZOOM} == -2 ) { $scalefactor = 2; }
    if ( $value{ZOOM} == -3 ) { $scalefactor = 3; }
    if ( $value{ZOOM} == -5 ) { $scalefactor = 5; }

    # Calculate coordinates of selected center point, if pt valid
    if ( $value{x} > 0 && $value{y} > 0 ) {
        if ( $value{proj} =~ /3/ ) {    # latlong projection
#$value{east} =~ s/_//g;
#$value{north} =~ s/_//g;
#$value{south} =~ s/_//g;
#$value{west} =~ s/_//g;
            $center{east} =
              ( ( $value{east} - $value{west} ) * ( $value{x} - 1 ) /
              $value{COLS} ) + $value{west};
            $center{east} = sprintf( "%10.4f", $center{east} );
            $center{north} = $value{north} -
              ( ( $value{north} - $value{south} ) * ( $value{y} - 1 ) /
              $value{ROWS} );
            $center{north} = sprintf( "%10.4f", $center{north} );
        }
        else {
#$value{east} =~ s/_//g;
#$value{north} =~ s/_//g;
#$value{south} =~ s/_//g;
#$value{west} =~ s/_//g;
            $center{east} =
              int( ( ( $value{east} - $value{west} ) * ( $value{x} - 1 ) /
              $value{COLS} ) + $value{west} );
            $center{north} =
              int( $value{north} -
              ( ( $value{north} - $value{south} ) * ( $value{y} - 1 ) /
              $value{ROWS} ) );
        }
    }

    #big if block to figure out if to QUERY, ZOOM, or PAN
    if ( $value{ZOOM} == 0 ) {    #query base maps
        $newMsg .= "<li>Location: ";
        $newMsg .= " <b>$center{east}E, $center{north}N </b>\n";
    #    $newMsg .= " <br><b>$value{east}E, $value{north}N </b>\n";

##################### RASTER QUERY SECTION ###################################
# start adding commented code for allowing multiple raster queries
#
#       my @rasts = @{ $Values{RASTER} };

        if ( !$value{RASTER} || $value{RASTER} eq "none" ) {

            #can't query a vector/site map
#$newMsg .= "<li><b>Sorry, no Raster basemap available to query</b></li>\n";

        }
        else {    # find category value for selected point
            &system_errors_logged("g.region save=saved_region");

        # main raster for loop to query values and display them
#        for ( $i = 0 ; $i <= $#vector_file ; $i++ ) {

            # reset region for raster query
            &system_silent("g.region raster=$value{RASTER}");



            # get category info
            $cmd =
"echo \"$center{east} $center{north}\"  | r.what -if input=$value{RASTER} null=\"null data\"";
            @catin = `$cmd`;
            @catdata = split ( /\|/, $catin[0] );
            $rastDesc = $RASTERS{ $value{RASTER} }->{descShort};
            $rastDesc = $value{RASTER} unless defined $rastDesc;
            $newMsg .= "<li>$rastDesc: <b>\n";
            $newMsg .= " ($catdata[3]) $catdata[4] </b></li>\n";
        
        if ( !$value{RASTER_SUB} || $value{RASTER_SUB} eq "none" ) {
            }
        else {
            # get category info
            $cmd =
"echo \"$center{east} $center{north}\"  | r.what -if input=$value{RASTER_SUB} null=\"null data\"";
            @catin = `$cmd`;
            @catdata = split ( /\|/, $catin[0] );
            $rastDesc = $RASTERS{ $value{RASTER_SUB} }->{descShort};
            $rastDesc = $value{RASTER_SUB} unless defined $rastDesc;
            $newMsg .= "<li>$rastDesc: <b>\n";
            $newMsg .= " ($catdata[3]) $catdata[4] </b></li>\n";
            }
            
            # restore saved region
            &system_errors_logged("g.region saved_region");

        }    #end of query on base rastermap

##################### VECTORS QUERY SECTION ###################################
       my @vects = @{ $Values{VECTOR} };
       # rebuild global vector form hash from hidden form vars

       if ( defined $value{SHOW_VECTORFORM} ) {
            @vectorforms = @{ $Values{SHOW_VECTORFORM} };
            foreach my $twopartform (@vectorforms) {
                my ( $part1, $part2 ) = split ( "xxx", $twopartform );
                #print ACCESSLOG "vectorforms: $part1 $part2\n";
                $vectorform{$part1} = $part2;
            }
        }

        @vector_file = @{ $Values{VECTORQUERY} };
        if ($#vector_file > 0) { # unfortunately v.what requires a monitor to be started (so use CELL)
           &system_errors_logged("export GRASS_HEIGHT=100; export GRASS_WIDTH=106");
           &system_silent("d.mon start=CELL");
        }

        # main vector for loop to query values and display them
        for ( $i = 0 ; $i <= $#vector_file ; $i++ ) {

            if ( defined $vectorform{ $vector_file[$i] } ) {
                $vectorforms = $vectorform{ $vector_file[$i] };
                #print ACCESSLOG "vectorforms2: $vectorforms\n";

            $vectInfo = $VECTORS{ $vects[$i] }->{descShort}; 
            #print ACCESSLOG "vectorquery: $vector_file[$i]  $center{east} $center{north}\n";
            $vector_detail = "";
            my $vector_detail_tmp = `$SCRIPT_PATH/lookitv.sh $center{east} $center{north} $vector_file[$i]`;
            # this is not supported by most webserver configs: 
            #my $vector_detail_tmp = `/home/httpd/html/glinks/progs/lookitv.sh $center{east} $center{north} $vector_file[$i]`;
            #my $vector_detail_tmp =  &system_errors_logged("/bin/echo $center{east} $center{north} | v.what -1 map=$vector_file[$i] 2>&1");

            # above shell script lookitv.sh has one line:
            #/bin/echo $1 $2 | v.what -1 map=$3 2>&1
            # which if done here would be:
            #$vector_detail = `echo $center{east} $center{north} | v.what -1 map=$vector_file[$i]\n`;
            # but which seems to need something a shell environment supplies 

            # clean up by igoring all but the Area Category attributes (not value)
            #$vector_detail .= $1 if ($vector_detail_tmp =~ /.*Area - Category \d+ (.*).*/msg);
            $vector_detail .= $1 if ($vector_detail_tmp =~ /.*Area - Category \d+ (.*)Size -.*/msg);
            #$vector_detail .= $vector_detail_tmp;

            #print ACCESSLOG "vectorquery2: $vector_detail\n";
            if ($vector_detail) { $newMsg .= "<li>$vectInfo: <b> $vector_detail </b></li>"; }
            }    #end of vector query if block
        }  #end vector for loop

        if ($#vector_file > 0) { # not sure need to worry about stopping the CELL monitor
           #&system_errors_logged("d.mon stop=CELL");
        }

################## SITES QUERY SECTION ########################################
        my @sites = @{ $Values{SITES} };
        # rebuild global site form hash from hidden form vars
 
        if ( defined $value{SHOW_SITEFORM} ) {
            @siteforms = @{ $Values{SHOW_SITEFORM} };
            foreach my $twopartform (@siteforms) {
                my ( $part1, $part2 ) = split ( "xxx", $twopartform );
                #print ACCESSLOG "siteforms: $part1 $part2\n";
                $siteform{$part1} = $part2;
            }
        }

        @sites_file = @{ $Values{SITEQUERY} };
        for ( $i = 0 ; $i <= $#sites_file ; $i++ ) {

            if ( defined $siteform{ $sites_file[$i] } ) {
                $siteforms = $siteform{ $sites_file[$i] };
                #print ACCESSLOG "siteforms2: $siteforms\n";
  
                $siteInfo = $SITES{ $sites[$i] }->{descShort};
                print ACCESSLOG "siteInfo: $siteInfo\n";

###########################################################################

                &get_siteforms($siteforms);

###########################################################################
                 }
              else { 
                 $siteforms = ""; 
                 $siteInfo = $SITES{ $sites[$i] }->{descShort};
                 }

           print ACCESSLOG "sitequery: $sites_file[$i]$center{east}$center{north}\n";
            # OK for GRASS 4.3 5.0pre1, not 5.0pre3
            #$sites_detail = `echo $center{east} $center{north} | s.what in=$sites_file[$i]`;
            # OK for GRASS 5.0pre3
            $center{east} =~ s/\s+//g;
            $center{north} =~ s/\s+//g;
            $sites_detail = `s.what -q map=$sites_file[$i] east_north=$center{east},$center{north}`;
            print ACCESSLOG "sitequery2: $sites_detail\n";

            if ($sites_detail) {
                # works 4.3 5.0pre1
                #( $siteinfo{e}, $siteinfo{n}, $siteinfo{rest} ) = split ( /\|/, $sites_detail );
                #$siteinfo{rest} =~ s/\"//g;
                # works 5.0pre3
                #( $siteinfo{e}, $siteinfo{n}, $siteinfo{rest} ) = split ( /\|/, $sites_detail );
                $siteinfo{rest} = $sites_detail;
                # this needs review
                $siteinfo{rest} =~ s/.*? in \w+//g;
                $siteinfo{rest} =~ s/.*\(E\)//g;
                $siteinfo{rest} =~ s/.*\(N\)//g;
                $siteinfo{rest} =~ s/\|/ /g;
           #     $siteinfo{rest} =~ s/.*?\dN//g;
                $siteinfo{rest} =~ s/.*?\d+:\d+.*?N//g;
                $siteinfo{rest} =~ s/"//g;
##               $siteinfo{rest} =~ s/.*\d+\|\d+(.*)/$1/g;
		@hotia = split(/ /,$siteinfo{rest});
		
                # need regexpression here:( $siteinfo{e}, $siteinfo{n}, $toss{rest}, $siteinfo{rest} ) = $siteinfo{rest};

                if ( $siteforms ) {
# siteforms wraps some HTML forms (stored in GL_forms.pl) 
# around the queried siteinfo
                    $newMsg .= "<li>$siteInfo: $sites_form <b> $siteinfo{rest} </b> $sites_endform</li>";
                }
                else {
                    $newMsg .= "<li>$siteInfo:<br><b> $siteinfo{rest} </b></li>";
		    $newMsg .= "<form><input type=\"button\" value=\"+ INFO\" ";
		    $newMsg .= "onClick=\"window.open('http://localhost/glinks/lookups.php?id=$hotia[4]','phpquery','width=800,height=200')\"></form>";

                }
            }
        }    #end of siteforms for loop

#########################################################

    }
    else {

        #do the appropriate pan or zoom calculation
        #which require creating a new gif
        $value{GIF} = "grasslinks.$$.gif";
        $value{PNG} = "grasslinks.$$.png";
        $value{PGM} = "grasslinks.$$.pgm";
        $value{PNM} = "grasslinks.$$.pnm";
        print ACCESSLOG
"zoom N: $value{north} S:$value{south} E:$value{east} W:$value{west} nw:$value{ns} ew:$value{ew} proj:$value{proj}\n";
        if ( $value{ZOOM} > 0 ) {    #shrink the region
            $adjust = ( 1 / $value{ZOOM} / 2 );
        }
        else {    #grow the region
            $adjust = $value{ZOOM} / -2;
        }
        print ACCESSLOG "LOGadjust $adjust\n";
        $north =
          $center{north} + ( ( $value{north} - $value{south} ) * $adjust );
        $south =
          $center{north} - ( ( $value{north} - $value{south} ) * $adjust );
        $east = $center{east} + ( ( $value{east} - $value{west} ) * $adjust );
        $west = $center{east} - ( ( $value{east} - $value{west} ) * $adjust );

        &system_errors_logged(
"g.region n=$north s=$south e=$east w=$west ewres=$value{ew} nsres=$value{ns}"
        );
        &get_coords;
        &adjust_region;
        &create_gif;
    }

    # Print out html
    #&begin_html( "Map Display, Zoom & Query", "FORM" );
    &begin_html( "Map Query", "FORM" );
    &create_display_html("$newMsg","");
    &end_ourform;
    &end_ourhtml;

}    # end of sub display3

#-------------------------------------------------------------------------
# NAME: report1 
#	Allows user to select one or two raster maps for area
#	calculation. User can also select region.
#	BUG - all rasters in $GLDATA_LIST appear in select boxes
#	but not all are suitable for area calculations (e.g. shaded-relief)
#	Also -importance of region and any masks (which aren't apparent
#	to user) needs to be conveyed to user
#-------------------------------------------------------------------------
sub report1 {

    &begin_html( "Areal Report of Mapped Information", "FORM" );

    print << "ENDPRINT";
	Areas are calculated as a total for each category within
	the selected raster map(s).  A report on "Counties" would give the total
	area for each county.  A report on "Counties" against "Wetlands to
	16 Categories" would show how much of each wetland type occurs
	in each county. 
	<br>
	<b>&#42;</b> View information on 
	<a href="$HTML_PATH/$GL_MATRIX" target="new">
	Raster Maps</a> available in GRASSLinks for Areal Calculations.
	<br> 
	<b>&#42;</b> Refer to the <a href="$GL_HELP#report" target="new">
	Help document</a> for further information.
	<br>
	<hr>
	<br>
	<li><b>Select a raster map for area calculation:</b>
	<br>
ENDPRINT

    &make_select( "RASTER", "4", "raster", "" );

    print '<br><li><b>
	Select an optional second raster map to overlay with the first:
	</b><br>
	';
    &make_select( "RASTER", "4", "raster", "" );

    print "<br><hr><br><li><b>Select a region:</b><br>\n";

    &make_select( "REGION", "4", "region", "" );

    print << "ENDPRINT";
	<br>
	<b>NOTE</b>: Your selected region setting is very
	important for Areal Reports. This is for two reasons: 1) the larger
	the region setting, the longer the process takes; and 2) the report
	analysis will be clipped to the selected GRASS region. For example,
	if you were to do "Counties" against "Wetlands" with a region
	setting of "Marin County", your area totals would be restricted to
	the predefined Marin County region box (which would include all of
	Marin and some of the surrounding counties).  
	<br>
	<hr><br>
ENDPRINT
    if ( $coord{proj} =~ /3/ ) {    # lat long
        print << "ENDPRINT";
	<li><b>Input a Raster Grid Resolution Setting between 0.01 and 0.001:</b>
	<INPUT type="text" size=5 name="RESOLUTION" value="0.001">
	degrees per Cell.
ENDPRINT
    }
    else {    # not lat long
        print << "ENDPRINT";
	<li><b>Input a Raster Grid Resolution Setting between 10 and 1000:</b>
	<INPUT type="text" size=5 name="RESOLUTION" value="250">
	Meters per Cell.
ENDPRINT
    }
    print << "ENDPRINT";
        <br>
	<b>NOTE:</b> 
	A coarse resolution will increase the speed of the reporting
	process but decrease the accuracy. But we suggest you only refine
	as needed. The finest resolution setting would be that of the
	original raster map(s).  See the appropriate metadata record
	for this value. 
	<br>
	<hr><br>
	<li><b>Select the units to report in:</b>
ENDPRINT

    &make_select( "UNITS", "1", "repUnits", "" );

    print "<br><hr><br>\n";
    &make_hidden( "ANALYSIS", "report2" );
    &make_hidden("PROJECT");
    &end_ourform( "Run Areal Report Calculation", "Reset" );
    print "(estimated run time 30 seconds)<BR>\n";
    &end_ourhtml;
}

#-------------------------------------------------------------------------
# NAME: report2
# Do the area tabulations based on the raster(s) selected in
# report1 and then return the formatted results to user
# BUG - area calculations trickiness includes: 1) resolution of
# the selected region (greater res. faster yet coarser results;
# and 2) masks. Masks determine the area of the calculation and
# need to be hand made and stored in the format:
# $GISDBASE/$LOCATION_NAME/grasslinks0/cell/MASK.<region>
# If no mask exists then the extent of the selected region is
# the area of the calculation. This stuff needs to be conveyed
# to user so they know what they are getting
#-------------------------------------------------------------------------
sub report2 {

    my ( $reportDesc, $rnumber, $rast1, $rast2, $regionInfo, @rasts );
    my ( $convert,    $units,   $att,   $cat,   $area );

    @rasts = @{ $Values{RASTER} };
    if ( !@rasts ) {    # if no rasters selected
        &error(
"Areal Report Operation Failed! <br>You must select at least one raster file!"
        );
        return;
    }

    if ( $rasts[1] ) {

        # two rasters selected - crosstab
        $rnumber = 2;
        $value{RASTER1} = $rasts[0];
        $value{RASTER2} = $rasts[1];
        &set_maxregion;

        # get some descriptive text
        $rast1      = $RASTERS{ $value{RASTER1} }->{descShort};
        $rast2      = $RASTERS{ $value{RASTER2} }->{descShort};
        $regionInfo = $REGIONS{ $value{REGION} }->{descShort};
        $reportDesc = "<li>Report for the map layer: ";
        $reportDesc .= "  <b>$rast1</b> <br><li>against the map layer: ";
        $reportDesc .= "  <b>$rast2</b> <br><li>within the region setting: ";
        $reportDesc .= "  <b>$regionInfo</b>";
    }
    else {

        # only one raster selected
        $rnumber = 1;
        $value{RASTER1} = $rasts[0];
        $value{RASTER}  = $value{RASTER1};
        &set_region();

        # get some descriptive text
        $rast1      = $RASTERS{ $value{RASTER1} }->{descShort};
        $regionInfo = $REGIONS{ $value{REGION} }->{descShort};
        $reportDesc = "<li>Report for the map layer: ";
        $reportDesc .= " <b>$rast1</b><br><li>within the region setting: ";
        $reportDesc .= " <b>$regionInfo</b>";
    }

    # reset the resolution to speed up process
    $value{RESOLUTION} = 250 unless defined $value{RESOLUTION};
    if ( $value{RESOLUTION} < 10 || $value{RESOLUTION} > 1000 ) {
        $value{RESOLUTION} = 250;
    }
    &system_silent("g.region res=$value{RESOLUTION}");
    $reportDesc .= "<br><li>Raster Grid Resolution Settings: ";
    $reportDesc .= " <b>$value{RESOLUTION} Meters per Cell </b>";

    #to the grass stats routine & write output to temp file
    &system_errors_logged(
"r.stats -aql fs=\"^\" input=$value{RASTER1},$value{RASTER2} > $TMP/stats.$$"
    );

    #create HTML code from GRASS temp stats file and close
    ###########################################################
    if ( $value{UNITS} eq "me" ) {
        $convert = 1;
        $units   = "sq meters";
    }
    elsif ( $value{UNITS} eq "k" ) {
        $convert = 0.000001;
        $units   = "sq kilometers";
    }
    elsif ( $value{UNITS} eq "mi" ) {
        $convert = 0.0000003861;
        $units   = "sq miles";
    }
    elsif ( $value{UNITS} eq "a" ) {
        $convert = 0.0002471;
        $units   = "acres";
    }
    elsif ( $value{UNITS} eq "h" ) {
        $convert = 0.0001;
        $units   = "hectares";
    }

    ###########################################################

    #print out Report HTML 
    &begin_html("Areal Calculations");

    print << "ENDPRINT";
	$reportDesc
	<br><hr><br>
ENDPRINT

    printf( "<PRE><B>%-44s %8s</B><BR>\n", "    Category:", "Area in $units" );
    open( STATS, "$TMP/stats.$$" );
    if ( $rnumber == 1 ) {    #only one raster
        while (<STATS>) {
            chop($_);
            ( $att, $cat, $area ) = split ( /\^/, $_ );
            printf( "%4s %-40s %12.2f<br>\n", $att, $cat, $area * $convert );
            if ( $. % 5 == 0 ) { print "<BR>\n"; }
        }    #end while loop
    }
    else {

        #doing a crosstab
        $old = "";

        while (<STATS>) {
            chop;
            ( $att1, $cat1, $att2, $cat2, $area ) = split (/\^/);
            if ( $att1 eq $old ) {
                printf( "    %4s %-30s %12.2f<br>\n", $att2, $cat2,
                    $area * $convert );
            }
            else {
                printf( "\n%4s %-30s<br>\n",          $att1, $cat1 );
                printf( "    %4s %-30s %12.2f<br>\n", $att2, $cat2,
                    $area * $convert );
            }
            $old = $att1;
        }    #end of while loop
    }    #end of if block

    close(STATS);
    print "</PRE>\n";

    print << "ENDPRINT";
	<br><hr>
	<i>Please refer to the <a href="$GL_HELP#report" target="new">
	Help document</a> for further information on interpreting these results.
	</i>
	<hr><br>
ENDPRINT

    &end_ourhtml;
    &flush(STDERR);
}    #end of report2

#-------------------------------------------------------------------------
# NAME: reclass1
# 	Displays html form for select map layers for reclass.	
#-------------------------------------------------------------------------

sub reclass1 {
    &begin_html( "Reclassification of Mapped Information", "FORM" );

    print << "ENDPRINT";
	This GIS tool allows you to create a new map from an
	existing one by combining categories from the chosen map into 
	one new category.  For example, one could create a map of East
	Bay counties by aggregating Contra Costa and Alameda from the 
	Counties map.  Or a user could simply isolate a category, 
	e.g., a map showing only Alameda County.
	<br>
	<b>&#42;</b> View information on 
	<a href="$HTML_PATH/$GL_MATRIX" target="new">
	Raster Maps</a> available in GRASSLinks for Reclass operations.
	<br> 
	<b>&#42;</b> Refer to the <a href="$GL_HELP#reclass" target="new">
	Help document</a> for further information.
	<br>
	<hr>
	<br>
	<li><b>Select a single <a href="$GL_HELP#raster" target="new">
	Raster</a> Map for reclassification:</b>
ENDPRINT

    &make_select( "RASTER", "4", "raster", "" );

    print "<br>\n";
    &make_hidden( "ANALYSIS", "reclass2" );
    &make_hidden("PROJECT");
    &end_ourform("Continue to Select Categories");
    &end_ourhtml;
}

#-------------------------------------------------------------------------
# NAME: reclass2
#	Displays selection list of unique attribute values (categories) for
#	the raster selected in reclass1 & allows user to choose which of these
#	values to select out or aggregate as a new category in a new map.
#-------------------------------------------------------------------------
sub reclass2 {

    my ( $numcats, $rasttmp );

    # make sure a valid raster selected
    if ( !$value{RASTER} || $value{RASTER} eq "none" ) {
        &error("No raster map selected");
        return;
    }

    $rasttmp = $RASTERS{ $value{RASTER} }->{descShort};
    $value{CURRENT_ATT} = 1;

    &begin_html( "Reclassification of Mapped Information, continued", "FORM" );

    print << "ENDPRINT";
	As an example, if you wanted to create a new counties
	map, the Map Filename might be "counties.sue" and the Data
	Label for category 1 could be "East Bay" and it might
	include Alameda and Contra Costa counties.  Category 2
	could be "North Bay" with Marin, Sonoma, Napa, and Solano
	counties.

	Refer to the <a href="$GL_HELP#reclass" target="new">Help
	document</a> for further information.
	<br>
	<b>1) Choose categories that you wish to aggregate from your 
	selected raster map: </b><br>$rasttmp
	<br>
	<li><b>Available Categories:</b><br>
ENDPRINT

    $numcats = &find_cats("$value{RASTER}");

    if ( $numcats < 1 ) {
        &error(
"Reclassification of map \"$rasttmp\" <br> failed due to insufficient number of categories in that map. <br>Please select a different map."
        );
        return;
    }

    &make_select_cats( "IN_CAT", "6", "1", "multiple" );
    print "<br>\n";

    print << "ENDPRINT";
	<li><b>Enter a Data Label for your aggregate category number 
	$value{CURRENT_ATT}:</b>
	<INPUT type="text" size=45 name="NEWCATLABEL">
	<BR><hr><BR>

	<b>2) Enter a new name for your map.</b><br>
	The reclass operation will create a new map.  A unique (but short)
	filename is necessary to avoid confusion with other users.
	<i>You must restrict punctuation to \"_\" and \".\", \
	do not include spaces. Any unapproved characters will be stripped.
	</i>
	<br>
	<b>Map Name:</b> <INPUT type="text" size=40 name="OUT_RASTER">
	<Br><hr><br>
	<b>3) Add more new categories or create the map:</b><br>
	You may opt to add more categories to your map, or if you have
	enough you can go on to create the requested map.
	<br>
	<INPUT TYPE="radio" NAME="NEXT" VALUE="MORE">
	Add more categories
	<b>or</b><br>
	<INPUT TYPE="radio" NAME="NEXT" VALUE="DONE" CHECKED>
	Create the new reclassified map
	<br>
ENDPRINT

    # this file must be defined here though not used until
    # reclass3 subroutine
    $value{RECLASS_RULES} = "reclassrules.$$";

    foreach $i ( "RASTER", "CURRENT_ATT", "RECLASS_RULES", "PROJECT" ) {
        &make_hidden($i);
    }
    &make_hidden( "ANALYSIS",                         "reclass3" );
    &end_ourform( "Continue Reclassification Operation", "Reset default values" );
    &end_ourhtml;
}

#-------------------------------------------------------------------------
# NAME: reclass3
# 	Loop for more reclass categories or run the GRASS reclass program
#	and return results in html to user.
# REQUIRES: THe file $value{RECLASS_RULES} must be pre-defined.
#-------------------------------------------------------------------------
sub reclass3 {

    my ( $theCats, $rasttmp );

    if ( !$value{OUT_RASTER} ) {
        &error("You must enter a name for the new Reclass Map!");
        return;
    }

    # First update reclass rules file
    @theCats = @{ $Values{IN_CAT} };
    if (@theCats) {    # if category(s) selected
        if ( !$value{NEWCATLABEL} ) {
            &error("No label input for the reclassification category");
            return;
        }
        else {
            open( RECLASS_RULES, ">>$TMP/$value{RECLASS_RULES}" );
            for ( $i = 0 ; $i <= $#theCats ; $i++ ) {
                print RECLASS_RULES
                  "$theCats[$i] = $value{CURRENT_ATT} $value{NEWCATLABEL}\n";
            }
            close(RECLASS_RULES);

        }
    }
    else {

        # no cats selected
        if ( $value{CURRENT_ATT} == 1 ) {
            &error("No category selected for Reclass Operation");
            return;
        }
    }

    #create form for further input or do the reclass and display
    if ( $value{NEXT} eq "MORE" ) {
        &find_cats("$value{RASTER}");
        $value{CURRENT_ATT}++;
        $rasttmp = $RASTERS{ $value{RASTER} }->{descShort};

        &begin_html( "Reclassified Map \"$value{OUT_RASTER}\"", "FORM" );

        print << "ENDPRINT";
		<b>1) Select the next categories from the map "$rasttmp" 
		to aggregate and include in your new map:</b>  
		$value{OUT_RASTER}
		<br>
		<li><b>Available Categories:</b><br>
ENDPRINT
        &make_select_cats( "IN_CAT", "6", "1", "multiple" );

        print << "ENDPRINT";
		<br>
		<li><b>Enter a Data Label for your aggregate category number
        	$value{CURRENT_ATT}:</b>
        	<INPUT type="text" size=45 name="NEWCATLABEL">
        	<BR><hr>
		<br>
		<b>3) Add more new categories or create the map:</b><br>
        	You may opt to add more categories to your map, or if you have
        	enough you can go on to create the requested map.
        	<br>
        	<INPUT TYPE="radio" NAME="NEXT" VALUE="MORE">
        	Add more categories
        	<br>or</b><br>
        	<INPUT TYPE="radio" NAME="NEXT" VALUE="DONE" CHECKED>
        	Create the new reclassified map
        	<br>
ENDPRINT

        foreach $i (
            "RASTER", "OUT_RASTER", "CURRENT_ATT", "RECLASS_RULES",
            "PROJECT"
          )
        {
            &make_hidden($i);
        }
        &make_hidden( "ANALYSIS", "reclass3" );
        &end_ourform("Continue Reclassification Operation");
        &end_ourhtml;

    }
    else {    #DONE -create the reclass
        &system_errors_logged(
"r.reclass input=$value{RASTER} output=$value{OUT_RASTER} < $TMP/$value{RECLASS_RULES}"
        );

        # create a GRASS colr file for the new reclass raster
        &system_errors_logged(
            "r.colors -wq map=$value{OUT_RASTER} color=rainbow");

        # Create gif reclass map
        $value{REGION} = "raster";
        &set_region;
        &get_coords;
        &adjust_region;
        &create_display_map;

        # Use a legend to convey reclass information
        $value{LEGEND} = "yes";

        # print out resultant display image
        &begin_html( "Reclassification of Mapped Information", "FORM" );
        &create_display_html("","");
        &end_ourform;
        &end_ourhtml;
    }    #end of if block
}

#-------------------------------------------------------------------------
# NAME: buffer1 
# 	Gives select form of available maps for buffer.
#-------------------------------------------------------------------------
sub buffer1 {

    &begin_html( "Buffering of Mapped Information", "FORM" );

    print << "ENDPRINT";
	This GIS tool allows you to create a new map
	by building concentric rings of known distance
	around categories in an existing map.  
	<p>
	<b>&#42;</b> View information on 
	<a href="$HTML_PATH/$GL_MATRIX" target="new">
	Raster Maps</a> available in GRASSLinks for buffering operations.
	<br> 
	<b>&#42;</b> Refer to the <a href="$GL_HELP#buffer" target="new">
	Help document</a> for further information.<br>
	<p>
	<hr>
	<p>
	<li><b>Select a single <a href="$GL_HELP#raster" target="new">
	Raster</a> Map for Buffer Operations:</b>
	<br>
ENDPRINT

    &make_select( "RASTER", "4", "raster", "" );
    print "<p>\n";

    &make_hidden( "ANALYSIS", "buffer2" );
    &make_hidden("PROJECT");
    &end_ourform("Continue Buffer Operation");
    &end_ourhtml;
}

#-------------------------------------------------------------------------
# NAME: buffer2 
# 	Select categories from raster selected in buffer1 around
# 	which to buffer; select buffer distance(s), units, region, 
#	resolution  and new map name. 
#-------------------------------------------------------------------------
sub buffer2 {

    my ( $numcats, $rasttmp );

    # make sure a raster selected
    if ( !$value{RASTER} ) {
        &error("No raster map selected!");
        return;
    }

    # get description of the selected raster map
    $rasttmp = $RASTERS{ $value{RASTER} }->{descShort};

    # begin creating html form
    &begin_html( "Buffering of Mapped Information, continued", "FORM" );

    print << "ENDPRINT";
	<b>1) Select categories</b> from the <b>$rasttmp</b> map 
	that you would like to create buffer rings around.
	<p>
	<li><b>Available Categories:</b><br>
ENDPRINT

    $numcats = &find_cats("$value{RASTER}");
    if ( $numcats < 1 ) {
        &error(
"Buffer Operation on map \"$rasttmp\" <br> failed due to insufficient number of categories in that map. <p>Please select a different map."
        );
        return;
    }

    &make_select_cats2( "CAT", "6", "1", "multiple" );

    print '<p><hr><p>
	<b>2) Select the units for your buffering distances:</b>
	 ';
    &make_select( "UNITS", "1", "bufUnits", "" );

    print << "ENDPRINT";
	<p>
	<b>3) Select the buffering distances.</b><br>
	Enter the outer distance for all rings separated by commas;
	e.g., 1,2,5 would create 3 rings at distances of 1, 2 and 5 units from the 
	edge of all categories selected above.
	<p>
	<b>Distances:</b> <INPUT type="text" size=40 name="DISTANCES">
	<p><hr><p>
	<b>4) Select a region:</b>
ENDPRINT
    &make_select( "REGION", "4", "region", "" );

    print << "ENDPRINT";
	<br>
	<i>NB: a smaller region will run faster.</i>
	<p>
	<b>5) Select a resolution to use:</b>
ENDPRINT
    &make_select( "RESOLUTION", "1", "resValues", "" );

    print << "ENDPRINT";
	<br>
	<i>NB: a coarse resolution will run much faster, but
	the pixels of your output raster file will be comparatively large.</i>
	<p>
	<hr>
	<p>
	<b>5) Enter a new name for your map.</b><br>
	The buffer operation will create a new map.  A unique (but short)
	filename is necessary to avoid confusion with other users.
	<i>You must restrict punctuation to alpha_numeric characters,
	any others will be deleted.</i>
	<p>
	<b>Your Map Name:</b>
	<INPUT type="text" size=40 name="OUT_RASTER">
	<p><hr><p>
ENDPRINT

    &make_hidden("RASTER");
    &make_hidden( "ANALYSIS", "buffer3" );
    &make_hidden("PROJECT");
    &end_ourform( "Create Buffer Map", "Reset" );
    &end_ourhtml;
}

#-------------------------------------------------------------------------
# NAME: buffer3
# 	Create a reclass rules file and then buffer  based on 
# 	parameters in this file.
#-------------------------------------------------------------------------
sub buffer3 {

    my ( $bufDesc, $catNo, $catName );
    $bufDesc = "<li>Buffer Ring(s) of <b>$value{DISTANCES} $value{UNITS}</b> ";
    $bufDesc .= " around the categories: ";

    # make sure buffer distances, and out_raster filename input
    if ( !$value{DISTANCES} ) {
        &error("No valid buffer distance(s) entered!");
        return;
    }
    if ( !$value{OUT_RASTER} ) {
        &error("No valid file name for new buffer map entered!");
        return;
    }

    #write to reclass rules file
    $value{RECLASS_RULES} = "$TMP/glbuffer.reclassrules.$$";

    @theCats = @{ $Values{CAT} };
    if (@theCats) {    # if cat(s) selected
        $bufDesc .= "<b> ";
        open( RECLASS_RULES, ">$value{RECLASS_RULES}" )
          || print STDERR
"GL3 buffer3 unable to open reclass rules file .$value{RECLASS_RULES}.\n";

        for ( $i = 0 ; $i <= $#theCats ; $i++ ) {
            ( $catNo, $catName ) = split ( /,/, $theCats[$i] );
            print RECLASS_RULES "$catNo = 1 origin of buffer\n";
            $bufDesc .= " $catName";
            if ( $i > 0 ) { $bufDesc .= " , "; }
        }

        $bufDesc .= "</b> ";
        close(RECLASS_RULES);

    }
    else {

        &error("No category/ies selected around which to buffer!");
        return;
    }

    #do the grass reclass prior to buffering
    &system_errors_logged(
"r.reclass input=$value{RASTER} output=buff.recl < $value{RECLASS_RULES}"
    );

    # set the region and resolution
    &set_region;
    &get_coords;
    &set_resolution;

    #do the buffering
    &system_errors_logged(
"r.buffer -q input=buff.recl output=$value{OUT_RASTER} distances=$value{DISTANCES} units=$value{UNITS}"
    );
    &system_silent("g.remove rast=buff.recl");

    &adjust_region;
    &create_display_map("$bufDesc");

    # return html file
    &begin_html( "Buffer of Mapped Information", "FORM" );
    &create_display_html("","");
    &end_ourform;
    &end_ourhtml;
    &flush(STDERR);
}

#-------------------------------------------------------------------------
# NAME: combine1 
# 	Select two raster maps for which attribute categories will
# 	then be displayed in selection lists;
# 	based on which of these are selected, a new map can be created.
#-------------------------------------------------------------------------
sub combine1 {

    &begin_html( "Coincidence of Mapped Information", "FORM" );

    print << "ENDPRINT";
	This GIS tool allows you to create a new map that highlights the
	overlap of information in two existing maps.
	For example, one could create a map of tidal wetlands in Marin County
	by choosing Marin from the Counties map and all tidal wetlands from the
	Wetlands to 16 categories map.
	<p>
	<b>&#42;</b> View information on 
	<a href="$HTML_PATH/$GL_MATRIX" target="new">
	Raster Maps</a> available in GRASSLinks for Combine operations.
	<br> 
	<b>&#42;</b> Refer to the <a href="$GL_HELP#combine" target="new">
	Help document</a> for further information.
	<p>
	<hr>
	<p>
	<li><b>Select two different raster maps to combine</b>
	<br>
ENDPRINT
    &make_select( "RASTER1", "4", "raster", "" );
    print "<p>";
    &make_select( "RASTER2", "4", "raster", "" );
    print "<p>\n";
    &make_hidden( "ANALYSIS", "combine2" );
    &make_hidden("PROJECT");
    &end_ourform("Continue Combine Operation");
    &end_ourhtml;
}

#-------------------------------------------------------------------------
# NAME: combine2 
# 	User selects categories to combine for rasters selected in
# 	combine1 form.
#-------------------------------------------------------------------------
sub combine2 {

    my ( $rast1, $rast2 );

    &begin_html( "Coincidence of Mapped Information, continued", "FORM" );

    if ( ( !( $value{RASTER1} && $value{RASTER2} ) )
        || $value{RASTER1} eq $value{RASTER2} )
    {
        &error("You need to select two different input maps");
        return;
    }
    else {    # proceed
        $rast1 = $RASTERS{ $value{RASTER1} }->{descShort};
        $rast2 = $RASTERS{ $value{RASTER2} }->{descShort};
        print << "ENDPRINT";
		<b>1) Select categories from each map that you would 
		like to combine.</b>
		<p>
		<dd><li><b>Categories for first map:</b> $rast1
ENDPRINT

        my $numcats = &find_cats("$value{RASTER1}");
        if ( $numcats < 1 ) {
            &error(
"Combine Operation on map \"$rast1\" <br> will fail due to insufficient number of categories in that map. <p>Please select a different map."
            );
            return;
        }
        &make_select_cats2( "CAT1", "4", "1", "multiple" );

        $numcats = &find_cats("$value{RASTER2}");
        if ( $numcats < 1 ) {
            &error(
"Combine Operation on map \"$rast2\" <br> will fail due to insufficient number of categories in that map. <p>Please select a different map."
            );
            return;
        }
        print "<p><dd><li><b>Categories for second map:</b> $rast2\n";
        &make_select_cats2( "CAT2", "4", "1", "multiple" );

        print "<p><hr><p>\n";
        print "<b>2) Select a region:</b>";

        &make_select( "REGION", "4", "region", "" );

        print "<br><i>NB: a smaller region will run faster.</i><p>\n";
        print "<b>3) Select a resolution to use:</b>";
        &make_select( "RESOLUTION", "1", "resValues", "" );

        print << "ENDPRINT";
		<br>
		<i>NB: a coarse resolution will run much faster, but
		the pixels of your output raster file will be very large.</i>
		<p>
		<hr>	
		<p>
		<b>4) Enter a new name for your map.</b><br>
		The combine operation will create a new map.  A unique (but short) 
		filename is necessary to avoid confusion with other users.
		<i>NB: you must restrict punctuation to alpha_numeric characters;
		any others will be deleted.</i>
		<p>
		<b>Map Name:</b>
ENDPRINT

        print "<INPUT type=\"text\" size=40 name=\"OUT_RASTER\">";
        print "<p>\n";

        &make_hidden( "ANALYSIS", "combine3" );
        &make_hidden("PROJECT");
        &make_hidden("RASTER1");
        &make_hidden("RASTER2");
        &end_ourform("Create the new Combine Map\n");
        &end_ourhtml;
    }    #endif
}    # end of sub combine2

#-------------------------------------------------------------------------
# NAME: combine3 
# 	Create and then return the new combined map request
#-------------------------------------------------------------------------
sub combine3 {

    my ( $combineDesc, $catNo, $catName );

    # make sure a mapname entered for new combine map
    if ( !$value{OUT_RASTER} ) {
        &error("You must enter a new name for the combine map!");
        return;
    }

    #get descriptive info on rasters selected for combine operation
    my $rast1       = $RASTERS{ $value{RASTER1} }->{descShort};
    my $rast2       = $RASTERS{ $value{RASTER2} }->{descShort};
    my $mapcalcFile = "$TMP/mapcalcrules.$$";

    #create mapcalc rules file needed by grass
    open( MAPCALC_RULES, "> $mapcalcFile" );
    print MAPCALC_RULES "$value{OUT_RASTER} = if (";

    #do rules for CAT1
    @theCats = @{ $Values{CAT1} };
    if (@theCats) {    # if cat(s) selected
        $combineDesc = "<li>Coincidence of the categories: <b>";
        for ( $i = 0 ; $i <= $#theCats ; $i++ ) {
            ( $catNo, $catName ) = split ( /,/, $theCats[$i] );
            print MAPCALC_RULES "$value{RASTER1} == $catNo ";
            $combineDesc .= " $catName";
            if ( $i < $#theCats ) {
                print MAPCALC_RULES "|| ";
                $combineDesc .= " , ";
            }
        }
        $combineDesc .= "</b> in the raster map <b>$value{RASTER1}</b>";
    }
    else {
        close(MAPCALC_RULES);
        &error("You must select category/ies for the map layer \"$rast1\" ");
        return;
    }

    print MAPCALC_RULES ", if ( ";

    #do rules for CAT2
    @theCats = @{ $Values{CAT2} };
    if (@theCats) {    # if cat(s) selected
        $combineDesc .= "<br>and the categories: <b>";
        for ( $i = 0 ; $i <= $#theCats ; $i++ ) {
            ( $catNo, $catName ) = split ( /,/, $theCats[$i] );
            print MAPCALC_RULES "$value{RASTER2} == $catNo ";
            $combineDesc .= " $catName";
            if ( $i < $#theCats ) {
                print MAPCALC_RULES "|| ";
                $combineDesc .= " , ";
            }
        }
        $combineDesc .= "</b> in the raster map: <b>$value{RASTER2}</b><br>";
    }
    else {
        close(MAPCALC_RULES);
        &error("You must select category/ies for the map layer \"$rast2\" ");
        return;
    }

    print MAPCALC_RULES ",1) )\n";
    close(MAPCALC_RULES);

    #set the region and resolution
    &set_maxregion;
    &get_coords;
    &set_resolution;

    #do the combine operation via mapcalc
    &system_errors_logged("r.mapcalc < $mapcalcFile");

    &adjust_region;
    &create_display_map("$combineDesc");

    #return html
    &begin_html( "Coincidence of Mapped Information", "FORM" );
    &create_display_html("","");
    &end_ourform;
    &end_ourhtml;
    &flush(STDERR);
}

#################################################################
# HTML formatting subroutines					# 
#################################################################

#-------------------------------------------------------------------------
# NAME: begin_html 
#	prints out html header
# ARGS: $_[0] = Title to appear in <title> section, and at top of <body>
#	$_[1] = set to "FORM" if html is a form
# REQUIRES: That the variable $project{title} indicates the
#	title of the current GL project to be noted on the html form.
# 	SHOULD BE CONFIGURED FOR LOCAL GRASSLinks Site.
#-------------------------------------------------------------------------
sub begin_html {
    my ( $title, $form ) = @_;

    print "Content-type: text/html\n\n";

    print << "ENDPRINT";
	<HTML>
	<HEAD>
	<TITLE>GRASSLinks: $title</TITLE>
	</HEAD>
	<BODY bgcolor=$color_bg1>

 <TABLE WIDTH=100%><TR><TD ALIGN=left>
        <FONT SIZE=+2 COLOR =green FACE="FIXED,LUCIDA,COURIER"><B>$GL_HEADER</B> </FONT>
        </TD><TD Align=right VALIGN=bottom>
        <FONT SIZE=+2 FACE="FIXED,LUCIDA,COURIER"><B>$project{title}</B></FONT>
        </TD></TR></TABLE>

ENDPRINT
	#<img src="$HTML_PATH/green_line.gif" width=100%>
    print "<FONT SIZE=+0> &nbsp;&nbsp;&nbsp;&nbsp; <B>$title:</B></FONT>" 
        unless $title eq "Map Query";

    if ( $form eq "FORM" && $value{SHOW_SITEFORM} eq "" ) {
        print "<FORM ACTION=\"$GL_ACTION\" METHOD=\"POST\">\n";
    }

}

#-------------------------------------------------------------------------
# NAME: end_ourform
#	Creates a submit and/or reset button on html form
# ARGS: $_[0] = Submit button text, (optional).
# ARGS: $_[1] = Reset button text, (optional).
#-------------------------------------------------------------------------
sub end_ourform {

    my ( $submit, $reset ) = @_;

    if ($submit) {
        print "<b><INPUT TYPE=\"submit\" VALUE=\"$submit\">\n";
    }
    if ($reset) {
        print " or <INPUT TYPE=\"reset\" VALUE=\"$reset\">\n";
    }
    print "</FORM>\n\n";
}

#-------------------------------------------------------------------------
# NAME: end_ourhtml 
# 	Print Links and text to appear at end of each GL html page
# 	SHOULD BE CONFIGURED FOR LOCAL GRASSLinks Site.
#-------------------------------------------------------------------------
sub end_ourhtml {

    &show_gladvanced;
#	<img src="$HTML_PATH/green_line.gif" width=100%>
    print << "ENDPRINT";
	<center>
	<sub>
	[
	<a href="$GL_URL">GRASSLinks Home</a>
	] [ 	
	<a href="$HTML_PATH/$GL_MATRIX">GRASSLinks Data</a>
	] [	
	<a href="$GL_HELP">GRASSLinks Help</a>
	] [	
	<a href="mailto: $GL_MAILTO ">GRASSLinks Feedback</a>
	]
	</sub>
	</center>
	</BODY>
	</HTML>
ENDPRINT

}

#-------------------------------------------------------------------------
# NAME: show_gladvanced 
# 	Prints out menu of advanced GRASSLinks options
#-------------------------------------------------------------------------
sub show_gladvanced {
	#<img src="$HTML_PATH/green_line.gif" width=100%>

    print << "ENDPRINT";
        <table border=0 cellpadding=0 bgcolor=$color_bg2 width=100%>
        <tr align=center>
	<td bgcolor=$color_bg1>&nbsp;</td>
        <td colspan=5>
        <a name="Advanced"><b>Advanced options - </b></a>
	<i>Select any of the following to perform real-time
	GIS operations.</i>
	</td></tr>
        <tr align=center>
        <td bgcolor=$color_bg1>&nbsp;</td>
        <td>
	<a href="$GL_ACTION?ANALYSIS=display1+PROJECT=$value{PROJECT}">
	Display</a></td>
        <td>
	<a href="$GL_ACTION?ANALYSIS=buffer1+PROJECT=$value{PROJECT}">
	Buffer</a></td>
        <td>
	<a href="$GL_ACTION?ANALYSIS=reclass1+PROJECT=$value{PROJECT}">
	Reclass</a></td>
        <td>
	<a href="$GL_ACTION?ANALYSIS=combine1+PROJECT=$value{PROJECT}">
	Combine</a></td>
        <td>
	<a href="$GL_ACTION?ANALYSIS=report1+PROJECT=$value{PROJECT}">
	Area Calculations</a></td>
        </tr>
ENDPRINT
# 3-9-02 this next section is redundant so dont show it
#    if ( -e "$HTML_FULLPATH/$GL_MATRIX" ) {
#        print << "ENDPRINT";
#		<tr align=center><td colspan=5>
#		<a href="$HTML_PATH/$GL_MATRIX" target="new">
#		View information on all GRASS data layers available via GRASSLinks
#		</a></td></tr>
#ENDPRINT
#    }

    print "</table>\n";

}

#-------------------------------------------------------------------------
# NAME: create_display_html
#	Creates html text for map display and a legend if requested.
# REQUIRES: $value{GIF} and $value{DISPLAY_DESC} files must exist
# ARGS: $_[0] = any query result text for display 
#	$_[1] = set to "NOTABLE" if utm coord table around gif image
#		not wanted. 
#-------------------------------------------------------------------------
sub create_display_html {

    my ( $qMesg, $wantTable ) = @_;
    print "<TABLE BORDER=0><TR><TD>&nbsp;</TD><TD BGCOLOR=$color_bg2>";

    # print map display description

    &print_file("$TMP/$value{DISPLAY_DESC}");

    # print query results
    if ($qMesg) {
        print "<FONT COLOR=GREEN><b>Query Results:</b></FONT>\n";
        print "<FONT SIZE=-1 COLOR=red>$qMesg</ul></FONT>\n";
    }

    #         print "</TD><TD>";
    &write_zoomtext;

    #         print "</TD></TR></TABLE>";
    print "</TD><TD BGCOLOR=$color_bg2>";

    # print out map gif in table or not
    if ( $wantTable eq "NOTABLE" ) {
        print << "ENDPRINT";
		<center>
        	<input type=image src="$HTML_TMP/$value{PNG}" ismap
        	height=$value{ROWS} width=$value{COLS}>
        	</center>
ENDPRINT

    }
    else {

        # update coord info if needed	
        my %display;
        if (%coord) {
               foreach $i ( "north", "south" ) {
                $value{$i} = $coord{$i};
                $display{$i} = sprintf( "%8.2f", $value{$i} );
            }
            foreach $i ( "east", "west" ) {
                $value{$i} = $coord{$i};
                $display{$i} = sprintf( "%8.1f", $value{$i} );
            }
        }
        else {    # need to reformat anyway for display purposes
                foreach $i ( "north", "south" ) {
                $display{$i} = sprintf( "%8.2f", $value{$i} );
            }
            foreach $i ( "east", "west" ) {
                $display{$i} = sprintf( "%8.1f", $value{$i} );
            }
        }
        my $fixed_title = $value{MAP_TITLE};
        $fixed_title =~ s/_/&nbsp;/g;
        print << "ENDPRINT";
		<center>
		<table border=0 CELLSPACING=1 CELLPADDING=1 >
		<tr align=center>
	        <td></td><td><B>$fixed_title</B></td></tr>
                <tr align=center>
                <td></td><td><BR></td></tr>
                <tr align=center><td></td>
	
		<td><FONT SIZE=-1><B>N</B> $display{north}</FONT></td></tr>
		<td></td>
		<tr>
		<td align=right><FONT SIZE=-1><B>W&nbsp;</B></FONT><FONT SIZE=-2><br> $display{west}</FONT></td>
		<td align=center>
       	 	<input type=image src="$HTML_TMP/$value{PNG}" ismap
       	 	height=$value{ROWS} width=$value{COLS}>
		</td>
		<td align=left><FONT SIZE=-1><B>&nbsp;E</B></FONT><FONT SIZE=-2><br> $display{east}</FONT></td></tr>
		<tr align=center>
		<td></td>
		<td><FONT SIZE=-1><B>S</B> $display{south}</FONT></td></tr>
		<td></td>
		</table>
        	</center>
		</td></tr></TABLE>
ENDPRINT
    }

    # legend no longer separate from the map so move call earlier
    # print out a legend table if user insists	
    #if( $value{LEGEND} eq "yes" ) {
    #	if ( $value{ANALYSIS} eq "display2" || $value{ANALYSIS} eq "reclass3" ) {
    #		&create_legend;
    #	} else {
    #		&print_file("$TMP/$value{LEGEND_TABLE}");
    #	}
    #}

    #	&write_zoomtext;
    &make_hidden_display;

}

#-------------------------------------------------------------------------
# NAME: make_select 
#	Create select menu based on array lists read in in the read_gldata
#	and set_gloptions subroutines.
# ARGS: $_[0] = Variable name in name value pair, (required).
#	$_[1] = The size of the select list box, (required).
#	$_[2] = The type of variable, needed to identify all options
#		in the $GLfiles{} hash, (required).
#	$_[3] = Set to "MULTIPLE" if multiple options can be selected. 
# NOTE: The order in which options are displayed in select lists 
#	depends on the order in which they appear in the GLDATA_LIST or
#	in the set_gloptions subroutine.
#-------------------------------------------------------------------------
sub make_select {

    my ( $varName, $listSize, $varType, $multy ) = ( "", "", "", "" );
    ( $varName, $listSize, $varType, $multy ) = @_;
    my ( $fname, $fdesc ) = ( "", "" );

    print "<SELECT $multy NAME=\"$varName\" SIZE=\"$listSize\">";

    for $newrec ( @{ $GLfiles{$varType} } ) {
        $fname = $newrec->[2];
        $fdesc = $newrec->[3];
        if ( $varType eq "region" ) {

            # set selected to $project{region}
            if ( $fname eq $project{region} ) {
                print "<OPTION VALUE=\"$fname\" SELECTED>$fdesc\n";
            }
            else {
                print "<OPTION VALUE=\"$fname\">$fdesc\n";
            }
        }
        else {

            # "!!" in field 1 indicates default value
            if ( $newrec->[1] =~ /\!\!/ ) {
                print "<OPTION VALUE=\"$fname\" SELECTED>$fdesc\n";
            }
            else {
                print "<OPTION VALUE=\"$fname\">$fdesc\n";
            }
        }
    }

    # Add any user created raster files to a raster pick list
    if ( $varType eq "raster" ) {
        for ( $i = 1 ; $i <= $N_GL_USERS ; $i++ ) {
            $loctmp = "$ENV{GISDBASE}/$ENV{LOCATION_NAME}/grasslinks$i";
# NEED to FIX: add fcell (floating point rasters) support;
# AND error log "ls: .. /grasslinks2/cell: No such file or directory"
            &system_errors_logged("ls $loctmp/cell > $TMP/$$.$i");
            open( CELL, "$TMP/$$.$i" );
            while (<CELL>) {
                print "<OPTION VALUE=$_\@grasslinks$i>$_\n";
            }
            close(CELL);
            unlink("$TMP/$$.$i");
        }
    }

    print "</SELECT>";
}

#-------------------------------------------------------------------------
# NAME: make_select_cats 
#	Create a select menu from an array @catlist created by sub find_cats
# ARGS: $_[0] = NAME, (required).
# 	$_[1] = SIZE of select list, (required).
#	#_[2] = default selected option, (optional). 
#	$_[3] = MULTIPLE, if multiple values can be selected, (optional).
# NOTE: This subroutine assumes the existance of the array @catlist
#	which contains the categories for the select list. This array
#	is created by the subroutine findcats().
#-------------------------------------------------------------------------
sub make_select_cats {

    my ( $varName, $listSize, $default, $multy ) = @_;
    my $counter = 1;

    print "<dd>\n";
    print "<SELECT $multy NAME=\"$varName\" SIZE=\"$listSize\">";

    foreach $i (@catlist) {
        if ( $default == $counter++ ) {
            print "<OPTION VALUE=\"$i\" SELECTED>$category{$i}\n";
        }
        else {
            print "<OPTION VALUE=\"$i\">$category{$i}\n";
        }
    }

    print "</SELECT>\n";
}

#-------------------------------------------------------------------------
# NAME: make_select_cats2
#	Create select menu from @catlist and %category created by sub find_cats
# ARGS: $_[0] = NAME, (required).
# 	$_[1] = SIZE of select list, (required).
#	#_[2] = default selected option, (optional). 
#	$_[3] = MULTIPLE, if multiple values can be selected, (optional).
# NOTE: This subroutine assumes the existance of the array @catlist
#	which contains the categories for the select list. This array
#	is created by the subroutine findcats().
#	THIS SUBROUTINE IS DIFFERENT FROM make_select_cats IN THAT IT
#	uses both the category number and name to form the VALUE.
#	See &buffer3 or &combine3 to see how it is used.
#-------------------------------------------------------------------------
sub make_select_cats2 {

    my ( $varName, $listSize, $default, $multy ) = @_;
    my $theValue;
    my $counter = 1;

    print "<dd>\n";
    print "<SELECT $multy NAME=\"$varName\" SIZE=\"$listSize\">";

    foreach $i (@catlist) {
        $theValue = "$i,$category{$i}";
        if ( $default == $counter++ ) {    #line no.=one to be selected
            print "<OPTION VALUE=\"$theValue\" SELECTED>$category{$i}\n";
        }
        else {
            print "<OPTION VALUE=\"$theValue\">$category{$i}\n";
        }
    }

    print "</SELECT>\n";
}

#-------------------------------------------------------------------------
# NAME: write_zoomtext 
#	Writes out html text for query, zoom and pan options
#	on the GL map display and query html form.
#-------------------------------------------------------------------------
sub write_zoomtext {

    my ( $thesite,   @sites );
    my ( $thevector, @vector );

    print << "ENDPRINT";
                <FONT COLOR=black SIZE=-1>
 <FORM ACTION=\"$GL_ACTION\" METHOD=\"POST\">
                <FONT COLOR=GREEN><B>Select Option:</b></FONT>
                (and click in map) 
                <LI><b>Zoom IN: </b>
		<INPUT TYPE="radio" NAME="ZOOM" VALUE="2"> 2
		<INPUT TYPE="radio" NAME="ZOOM" VALUE="3"> 3
		<INPUT TYPE="radio" NAME="ZOOM" VALUE="5"> 5
		<BR>
		<b>OUT: </b>
		<INPUT TYPE="radio" NAME="ZOOM" VALUE="-2"> 2
		<INPUT TYPE="radio" NAME="ZOOM" VALUE="-3"> 3
		<INPUT TYPE="radio" NAME="ZOOM" VALUE="-5"> 5
                <LI><INPUT TYPE="radio" NAME="ZOOM" VALUE="1"> 
                <b>Pan</b> (click new center)
                <LI><INPUT TYPE="radio" NAME="ZOOM" VALUE="0" CHECKED> 
                <b>Query</b> location and map values
ENDPRINT

    if ( defined $value{SITES} ) {
# 3-9-02 Ia stop showing list; will just make hidden
#        print << "ENDPRINT";
#		<BR><b>Site Query: </b><BR>
#		<SELECT NAME="SITEQUERY" SIZE="2" MULTIPLE>
#ENDPRINT
        @sites = @{ $Values{SITES} };
        if (@sites) {    # if site(s) selected
            for ( $i = 0 ; $i <= $#sites ; $i++ ) {
                last if $sites[$i] eq "none";
                $thesite = $sites[$i];
                print
# 3-9-02 IIa
#"<OPTION VALUE=\"$thesite\" SELECTED>$SITES{ $sites[$i] }->{descShort}\n";
"<INPUT TYPE=\"hidden\" NAME=\"SITEQUERY\" VALUE=\"$thesite\" SELECTED>\n";
            }    # end for
            # 3-9-02 IIIa
            #print "<OPTION VALUE=\"none\"> None</SELECT>\n";
        }

        #print "</UL>";
    }

    if ( defined $value{VECTOR} ) {    # new vector query results
# 3-9-02 Ib stop showing list; will just make hidden
#        print << "ENDPRINT";
#		<BR><b>Vector Query: </b><BR>
#		<SELECT NAME="VECTORQUERY" SIZE="2" MULTIPLE>
#ENDPRINT
        @vects = @{ $Values{VECTOR} };
        if (@vects) {    # if vect(s) selected
            for ( $i = 0 ; $i <= $#vects ; $i++ ) {
                last if $vects[$i] eq "none";
                $thevect = $vects[$i];
                print
# 3-9-02 IIb
#"<OPTION VALUE=\"$thevect\" SELECTED>$VECTORS{ $vects[$i] }->{descShort}\n";
"<INPUT TYPE=\"hidden\" NAME=\"VECTORQUERY\" VALUE=\"$thevect\" SELECTED>\n";
            }    # end for
            # 3-9-02 IIIb
            #print "<OPTION VALUE=\"none\"> None</SELECT>\n";
        }
        print "</UL>";
    }

}

#-------------------------------------------------------------------------
# NAME: make_hidden
#	Writes hidden variables to html file.
# ARGS: $_[0] = NAME of hidden variable, (required)
#	$_[1] = Value for hidden variabl, (optional - if not set, this
#		value is set to $value{NAME} 
#-------------------------------------------------------------------------
sub make_hidden {
    my $varName;
    if ( $_[0] ) {
         $varName = $_[0];
    }

    my $varValue;
    if ( $_[1] ) {
        $varValue = $_[1];
    }
    else {
        $varValue = $value{$varName};
    }

    print "<INPUT TYPE=\"hidden\" NAME=\"$varName\" VALUE=\"$varValue\">\n";

}

#-------------------------------------------------------------------------
# NAME: make_hidden_display 
#	Write hidden variables for display3 subroutine.
#-------------------------------------------------------------------------
sub make_hidden_display {

    my ( @a, @sites, @siteforms, @vectors, @vectorforms );

    @a = (
        "GIF",          "PNG", "REGION", "SITELABELS", "PROJECT",
        "GIFSIZE",      "COLS",   "ROWS",       "PMAP_SCRIPT",
        "DISPLAY_DESC", "LEGEND", "LEGEND_TABLE", "MAP_TITLE", "SUB_OPACITY"
    );

    foreach $i (@a) {
        &make_hidden($i);
    }

    foreach $i ( "proj", "north", "south", "east", "west", "ew", "ns" ) {

        # note: if no value for %coord hash, then set to
        # $value{north}, etc in sub make_hidden
# NEW 4/22/02 
# remove spaces from all coordinates - fixes UTM bug
# otherwise they get converted to underscores which we dont want
        $coord{$i} =~ s/\s+//g;
        &make_hidden( "$i", "$coord{$i}" );
    }

    @sites = @{ $Values{SITES} };
    for ( $i = 0 ; $i <= $#sites ; $i++ ) {
        &make_hidden( "SITES", $sites[$i] );
    }

    # new section for site forms - may be redundant
    if ( defined $value{SHOW_SITEFORM} ) {
        @siteforms = @{ $Values{SHOW_SITEFORM} };
        if (@siteforms) {
            for ( $i = 0 ; $i <= $#siteforms ; $i++ ) {

                #my $siteplusform = $sites[$i] . "xxx" . $siteforms[$i];
                print ACCESSLOG "siteplusform: $siteforms[$i] \n";
                &make_hidden( "SHOW_SITEFORM", $siteforms[$i] );
            }
        }
    }
    else {    # first time need to set from global not hidden form vars
        foreach my $thissiteform ( keys(%siteform) ) {
            my $mystring = $thissiteform . "xxx" . $siteform{$thissiteform};
            &make_hidden( "SHOW_SITEFORM", $mystring );
        }
    }

    # new for vector query
    @vectors = @{ $Values{VECTOR} };
    for ( $i = 0 ; $i <= $#vectors ; $i++ ) {
        &make_hidden( "VECTOR", $vectors[$i] );
    }

    # new section for vector forms - may be redundant
    if ( defined $value{SHOW_VECTORFORM} ) {
        @vectorforms = @{ $Values{SHOW_VECTORFORM} };
        if (@vectorforms) {
            for ( $i = 0 ; $i <= $#vectorforms ; $i++ ) {

                #my $siteplusform = $sites[$i] . "xxx" . $siteforms[$i];
                print ACCESSLOG "vectorplusform: $vectorforms[$i] \n";
                &make_hidden( "SHOW_VECTORFORM", $vectorforms[$i] );
            }
        }
    }
    else {    # first time need to set from global not hidden form vars
        foreach my $thisvectorform ( keys(%vectorform) ) {
            my $mystring = $thisvectorform . "xxx" . $vectorform{$thisvectorform};
            &make_hidden( "SHOW_VECTORFORM", $mystring );
        }
    }
# new exception - need to generalize this
    if ( defined $value{species} ) {
       &make_hidden( "species", "$value{species}" );
       }

    &make_hidden( "ANALYSIS", "display3" );
    &make_hidden( "RASTER",   "$value{OUT_RASTER}" );
    &make_hidden( "RASTER_SUB",   "$value{OUT_RASTER_SUB}" );

}

#################################################################
# GRASS Helper Subroutines					#
#################################################################
#-------------------------------------------------------------------------
# NAME: get_filepath 
# 	Get path and filename of a GRASS map element file such as
# 	category, color, metadata for a grass raster file.
# ARGS: $_[0] = grass element for which to find file, (required).
#	$_[1] = grass raster file for which filename is sought. (required).
# RETURNS: The name of the file
#-------------------------------------------------------------------------
sub get_filepath {

    my ( $grassElement, $grassFile ) = @_;
    my ( @junky,        $junk );

    @junky = `g.findfile element=$grassElement file=$grassFile`;
    ( $junk, $filename ) = split ( /=/, $junky[3] );
    $filename =~ s/'//g;

    return $filename;
}

#-------------------------------------------------------------------------
# NAME: find_cats 
#	Gets the unique category values for a grass raster file and
#	sticks them in array @catlist.
# ARGS: $_[0] = grass raster file
# RETURNS: Implicitly returns , the array @catlist - sets value for it.
#	Explicitly returns $totalcats, the number of categories.
#-------------------------------------------------------------------------
sub find_cats {

    my $raster = $_[0];
    my $theCatsFile;
    @catlist   = ();
    $totalcats = 0;

    $theCatsFile = &get_filepath( "cats", $raster );

    open( CATS, $theCatsFile )
      || print STDERR "GL find_cats can't open cats file for $raster.\n";
    while (<CATS>) {
        chop;
        @data = split (/:/);
        if ( $data[0] && $data[1] ) {
            $category{ $data[0] } = $data[1];
            @catlist = ( @catlist, $data[0] );
            $totalcats++;
        }
    }
    close(CATS);
    return $totalcats;
}

#-------------------------------------------------------------------------
# NAME: set_region 
# 	Sets grass region by which grass operations will be bound
# 	for grass operations that only use one raster at a time.
#-------------------------------------------------------------------------
sub set_region {

    if ( $value{REGION} eq "raster" ) {
        if ( $value{RASTER} && $value{RASTER} ne "none" ) {
            &system_errors_logged("g.region raster=$value{RASTER}");
        }
        else {
            &system_errors_logged("g.region $DEFAULT_REGION");
        }
    }
    else {
        if ( $value{REGION} && $value{REGION} ne "none" ) {

            # set the region to one defined by user
            &system_errors_logged("g.region $value{REGION}");
        }
        else {
            &system_errors_logged("g.region $DEFAULT_REGION");
        }
    }

}

#-------------------------------------------------------------------------
# NAME: set_maxregion 
#	Set region when more than one raster has been selected
#	to largest extent of the two rasters or to a specifically
#	set region in value{REGION}.
#-------------------------------------------------------------------------
sub set_maxregion {

    my %coord1;
    my %coord2;

    if ( defined $value{REGION} && $value{REGION} ne "raster" ) {
        &system_errors_logged("g.region $value{REGION}");
    }
    else {

        #set region to first raster, then get its extent coordinates
        &system_errors_logged("g.region raster=$value{RASTER1}");
        &get_coords;
        %coord1 = %coord;

        #set region to second raster, then get its extent coordinates	
        &system_errors_logged("g.region raster=$value{RASTER2}");
        &get_coords;
        %coord2 = %coord;

        foreach $i ( "north", "east" ) {
            if ( $coord2{$i} < $coord1{$i} ) {
                $coord{$i} = $coord1{$i};
            }
            else {
                $coord{$i} = $coord2{$i};
            }
        }
        foreach $i ( "south", "west" ) {
            if ( $coord2{$i} > $coord1{$i} ) {
                $coord{$i} = $coord1{$i};
            }
            else {
                $coord{$i} = $coord2{$i};
            }
        }
        foreach $i ( "ew", "ns" ) {
            if ( $coord2{$i} < $coord1{$i} ) {
                $coord{$i} = $coord1{$i};
            }
            else {
                $coord{$i} = $coord2{$i};
            }
        }
        &system_errors_logged( "g.region n=$coord{north} s=$coord{south}\
		e=$coord{east} w=$coord{west} ewres=$coord{ew} nsres=$coord{ns}"
        );
    }
}

#-------------------------------------------------------------------------
# NEW routine for latlong
# NAME: coordconvert
#       Convert degrees:minutes<:seconds> to decimal degrees
#-------------------------------------------------------------------------
sub coordconvert {
    %signs = (
        ''  => 1.0,
        ' ' => 1.0,
        'N' => 1.0,
        'S' => -1.0,
        'E' => 1.0,
        'W' => -1.0,
        'n' => 1.0,
        's' => -1.0,
        'e' => 1.0,
        'w' => -1.0
    );

    my ($c) = @_;
    my ( $s, $deg );

    $s = 1;
    $c =~ s/([NSEW])//;
    if ( $signs{$1} ) { $s = $signs{$1} }

    if ( $c =~ /(.*):(.*):(.*)/ ) {    # have degrees:minutes:seconds
        $deg = $s * ( $1 + $2 / 60 + $3 / 3600 );
    }
    elsif ( $c =~ /(.*):(.*)/ ) {    # have degrees:minutes
        $deg = $s * ( $1 + $2 / 60 );
    }
    else {
        $c =~ /(.*)/;    # have degrees
        $deg = $s * ($1);
    }
    return $deg;
}

#-------------------------------------------------------------------------
# NAME: get_coords 
#	Gets the coord values for the current GRASS WIND file.
#	Puts these coords in the global/static  hash %coord{}.
#-------------------------------------------------------------------------
sub get_coords {

    my ( $name, $value );
    %coord = ();
    open( WIND, "$LOCATION/WIND" )
      || print STDERR
      "GL get_coords couldn't open WIND file = .$LOCATION/WIND.\n";

    while (<WIND>) {

        # debug
        #        print ACCESSLOG "WIND $_\n";
        #s/[:-]//g;
        s/[-]//;
        s/ resol//g;
        s/\t/ /g;
        s/_//g;
        chop;

        #($name, $value) = split(/ +/);
        if (/(.*?):(.*)/) { ( $name, $value ) = ( $1, $2 ); }

                print ACCESSLOG "name value $name $value\n";
        if ( $value =~ /[\:NSEWnsew]/ && $coord{proj} =~ /3/) {
            $coord{$name} = coordconvert($value);
        }
        else { $coord{$name} = $value; }

        #        print ACCESSLOG "new $name $coord{$name}\n";
    }
    close(WIND);
    &flush(STDERR);
}

#-------------------------------------------------------------------------
# NAME: adjust_region 
#	Figure out scaled dimensions of the image, based on region & gifsize.
#-------------------------------------------------------------------------
sub adjust_region {

    if ( !$value{GIFSIZE} ) { $value{GIFSIZE} = 400; }

    if ( $coord{rows} > $coord{cols} ) {

        #portrait orientation
        $value{ROWS} = $value{GIFSIZE};
        $value{COLS} = int( $value{GIFSIZE} * $coord{cols} / $coord{rows} );
    }
    else {

        #landscape orientation
        $value{COLS} = $value{GIFSIZE};
        $value{ROWS} = int( $value{GIFSIZE} * $coord{rows} / $coord{cols} );
    }

    # now adjust resolution of region based on target image
    # FORMERLY:
    # $save_ew = $coord{ew};
    # $save_ns = $coord{ns};
    # &system_errors_logged("g.region nsres=$save_ns ewres=$save_ew");

    # changed to Simon Cox's formula by plf 1/14/97
    $coord{ns} = ( $coord{north} - $coord{south} ) / $value{ROWS};
    $coord{ew} = ( $coord{east} - $coord{west} ) / $value{COLS};

#        print ACCESSLOG "adjust N:$coord{north} S:$coord{south}  E:$coord{east} W: $coord{west} ns $coord{ns} ew $coord{ew} proj $coord{proj}\n";
    &system_errors_logged("g.region nsres=$coord{ns} ewres=$coord{ew}");

}

#-------------------------------------------------------------------------
# NAME: set_resolution 
#	Resets region based on a resolution selection.
#-------------------------------------------------------------------------
sub set_resolution {

    # set default res to coarse (5)
    if ( $coord{proj} =~ /3/ ) {    # we are in latlong
        print ACCESSLOG "proj: $coord{proj}\n";
        $value{RESOLUTION} = 0.003 unless defined $value{RESOLUTION};

        # min base resolution is 0.01 unless user wants fine
        unless ( $value{RESOLUTION} == 1 ) {
            if ( ( $coord{ew} + $coord{ns} ) < 0.006 ) {
                $coord{ew} = 0.003;
                $coord{ns} = 0.003;
            }
        }
    }
    else {    # not in latlong, assume UTM or similar
        $value{RESOLUTION} = 5 unless defined $value{RESOLUTION};

        # min base resolution is 100 unless user wants fine
        unless ( $value{RESOLUTION} == 1 ) {
            if ( ( $coord{ew} + $coord{ns} ) < 190 ) {
                $coord{ew} = 100;
                $coord{ns} = 100;
            }
        }
    }

    $coord{ew} = int( $coord{ew} * $value{RESOLUTION} );
    $coord{ns} = int( $coord{ns} * $value{RESOLUTION} );

    &system_errors_logged("g.region ewres=$coord{ew} nsres=$coord{ns}");
}

#################################################################
# SUBROUTINES FOR CREATING GRASSLinks GIF Images 		#	
#################################################################

#-------------------------------------------------------------------------
# NAME: make_pmap 
# 	Create the pmap script needed by grass to create map gif.
# REQUIRES: $value{PMAP_SCRIPT} already defined by calling function.
# RETURNS:  $dispMsg containing text info about display map.
#-------------------------------------------------------------------------
sub make_pmap {

    $dispMsg = "<FONT SIZE=-1>";
    my (
        $rastDesc, $subrastDesc, @vects,    $num,    $vcolor, $vectInfo,
        @sites,    $siteInfo, $scolor, $ssize
    );

    open( PMAP_SCRIPT, ">$TMP/$value{PMAP_SCRIPT}" );

    #write out raster information to PMAP scripts
    if ( $value{RASTER} ne "none" && $value{RASTER} ) {
        print PMAP_SCRIPT "raster $value{RASTER}\n";

        # may add choice here - I like approx for non dithered look
        #print PMAP_SCRIPT "    colormode best\n";
        print PMAP_SCRIPT "    colormode approx\n";

        #print PMAP_SCRIPT "    barscale 45% 5%\n";
        #print PMAP_SCRIPT "    unit km\n";
        #print PMAP_SCRIPT "    length 10\n";
        #print PMAP_SCRIPT "    width 1\n";
        #print PMAP_SCRIPT "    style tick\n";
        #my $textfactor = int ($scalefactor * 1000);
        #print PMAP_SCRIPT "    textsize $textfactor\n";
        #print PMAP_SCRIPT "    textfont romanc\n";
        #print PMAP_SCRIPT "    background white\n";
        #print PMAP_SCRIPT "    end\n";
        #print PMAP_SCRIPT "colortable y\n";
        $rastDesc = $RASTERS{ $value{RASTER} }->{descShort};
        $rastDesc = $value{RASTER} unless defined $rastDesc;
        $dispMsg .= "\n<FONT SIZE=-1><li>Basemap: <b>$rastDesc</b></li></FONT>";
    }
    if ( $value{RASTER_SUB} ne "none" && $value{RASTER_SUB} ) {
        $subrastDesc = $RASTERS{ $value{RASTER_SUB} }->{descShort};
        $subrastDesc = $value{RASTER_SUB} unless defined $subrastDesc;
        #my $sub_op_label = $opacity[2][3];
        my $sub_op_label = "unknown";
        if ($value{SUB_OPACITY} == 2.0) { $sub_op_label = "variable";}
        elsif ($value{SUB_OPACITY} == 1.0) { $sub_op_label = "zero";}
        elsif ($value{SUB_OPACITY} > 0.74) { $sub_op_label = "minimal";}
        elsif ($value{SUB_OPACITY} > 0.59) { $sub_op_label = "low";}
        elsif ($value{SUB_OPACITY} > 0.49) { $sub_op_label = "medium";}
        elsif ($value{SUB_OPACITY} > 0.39) { $sub_op_label = "med. high";}
        elsif ($value{SUB_OPACITY} > 0.24) { $sub_op_label = "v. high";}
        elsif ($value{SUB_OPACITY} == 0.0) { $sub_op_label = "total";}
        else { $sub_op_label = "unknown";}
       
        #$subrastDesc .= " at " . $value{SUB_OPACITY} if ($value{SUB_OPACITY});
        $subrastDesc .= " with " . $sub_op_label . " opacity" if ($value{SUB_OPACITY});
        $dispMsg .= "\n<FONT SIZE=-1><li>Overlay: <b>$subrastDesc</b></li></FONT>";
    }
    # write out vector information if vects selected 
    @vects = @{ $Values{VECTOR} };
    if (@vects) {
        for ( $i = 0 ; $i <= $#vects ; $i++ ) {
            last if $vects[$i] eq "none";
            $num    = $i + 1;              # because of numbering
            $vcolor = "VCOLOR" . $num;
            $vcolor = $value{$vcolor} or $vcolor = $DEFAULT_COLOR;
            my ($vbgcolor, $vbgcolor2);
            if ($vcolor eq "white" || $vcolor eq "yellow") {   
                # Need to contrast light colors better so use a table + bgcolor
                $vbgcolor = "<table><tr><td bgcolor=brown>";
                $vbgcolor2 = "</td></tr></table>";
               }
            else { $vbgcolor = ""; $vbgcolor2 = ""; }
 
            print PMAP_SCRIPT "vector $vects[$i]\n";
            print PMAP_SCRIPT "    color $vcolor\n";

            #set vector width to 1
if ($vcolor eq "yellow") {
            print PMAP_SCRIPT "    width 4\n";
}
else {
            print PMAP_SCRIPT "    width 1\n";

}
            #print PMAP_SCRIPT "    style 0011\n";
            print PMAP_SCRIPT "    end\n";
            $vectInfo = $VECTORS{ $vects[$i] }->{descShort};
            print ACCESSLOG "$VECTORS{ $vects[$i] }->{descShort}\n";
            print ACCESSLOG "$VECTORS{ $vects[$i] }->{vectorform}\n";
            $vectorform{ $vects[$i] } = $VECTORS{ $vects[$i] }->{vectorform};
            print ACCESSLOG "vectorform: $vects[$i] => $vectorform{$vects[$i]}\n";

            # new add siteform info always a hidden variable so
            # goes to html display not to pmap script
#$siteInfo .= "\n<INPUT TYPE=\"hidden\" NAME=\"SHOW_VECTORFORM\" VALUE=\"$vectorform[$i]\">\n";
            $dispMsg .=
              "\n<FONT SIZE=-1><li><b>$vectInfo</b> in </FONT>$vbgcolor <FONT SIZE=-1 COLOR=$vcolor>$vcolor</FONT>$vbgcolor2</li>";
        }
    }

    # write out sites information if sites selected
    @sites = @{ $Values{SITES} };
    if (@sites) {
        for ( $i = 0 ; $i <= $#sites ; $i++ ) {
            last if $sites[$i] eq "none";
            $num    = $i + 1;              # because of numbering
            $scolor = "SCOLOR" . "$num";
            $scolor = $value{$scolor} or $scolor = $DEFAULT_COLOR;
            my ($sbgcolor, $sbgcolor2);
            if ($scolor eq "white" || $scolor eq "yellow") { 
                $sbgcolor = "<table><tr><td bgcolor=brown>";
                $sbgcolor2 = "</td></tr></table>";
               } 
            else { $sbgcolor = ""; $sbgcolor2 = ""; }
            $ssize  = "SSIZE" . "$num";
            $ssize = $value{$ssize} or $ssize = $DEFAULT_SSIZE;
            
            print PMAP_SCRIPT "sites $sites[$i]\n";
            print PMAP_SCRIPT "    color $scolor\n";

            if ( $ssize =~ /\d+/ ) {       # a number for size no icon
                print PMAP_SCRIPT "    size $ssize\n";
            }
            else {    # use an icon created with p.icons
                print PMAP_SCRIPT "    size 1\n";
                print PMAP_SCRIPT "    icon $ssize\n";
            }
            print ACCESSLOG "SITELABELS= $value{SITELABELS}\n";
            if ( $value{SITELABELS} ne "no" ) {
                print PMAP_SCRIPT "    desc\n";
            }
            print PMAP_SCRIPT "    end\n";

            #print PMAP_SCRIPT "labels $sites\n";
            $siteInfo = $SITES{ $sites[$i] }->{descShort};
            print ACCESSLOG "$SITES{ $sites[$i] }->{descShort}\n";
            print ACCESSLOG "$SITES{ $sites[$i] }->{siteform}\n";
            $siteform{ $sites[$i] } = $SITES{ $sites[$i] }->{siteform};
            print ACCESSLOG "siteform: $sites[$i] => $siteform{$sites[$i]}\n";

            # new add siteform info always a hidden variable so
            # goes to html display not to pmap script
#$siteInfo .= "\n<INPUT TYPE=\"hidden\" NAME=\"SHOW_SITEFORM\" VALUE=\"$siteform[$i]\">\n";
            $dispMsg .=
"\n<FONT SIZE=-1><li>Sites: <b>$siteInfo</b> in </FONT>$sbgcolor <FONT SIZE=-1 COLOR=$scolor> $scolor </FONT>$sbgcolor2</li>";
        }
    }

    # close up the scripts
    #print PMAP_SCRIPT "end\n";
    close(PMAP_SCRIPT);

    return $dispMsg;

} # end make_pmap

#-------------------------------------------------------------------------
# NAME: create_gif 
#	Creates a gif in file $value{GIF} based on a valid PMAP_SCRIPT. 
# REQUIRES: PMAP_SCRIPT script ALREADY created as $value{PMAP_SCRIPT}, 
#	$value{GIF} and $value{GIFSIZE} already defined by calling subroutine.
#-------------------------------------------------------------------------
sub create_gif {

    $ENV{WIDTH}  = $value{GIFSIZE};
    $ENV{HEIGHT} = $value{GIFSIZE};

    ###########
    # move create_legend call to here (now that legend is back within p.map.new
    # and no longer separate
    # print out a legend table if user insists	
    if ( $value{LEGEND} ne "no" ) {
        if ( $value{ANALYSIS} eq "display2" || $value{ANALYSIS} eq "reclass3" )
        {
            &create_legend;
        }
        else {

            # this no longer expected to work
            #&print_file("$TMP/$value{LEGEND_TABLE}");
        }
    }

    #
    ###########

    $value{PMAP_SUB_SCRIPT} = "pmap_sub_script.$$";
  open( PMAP_SUB_SCRIPT, ">$TMP/$value{PMAP_SUB_SCRIPT}" )
      || print STDERR
"Unable to open pmap_sub script $TMP/$value{PMAP_SUB_SCRIPT} in sub create_display_map\n";
    print PMAP_SUB_SCRIPT "raster $value{RASTER_SUB}\n end\n";
    close(PMAP_SUB_SCRIPT);


    # NEW create our SUB raster layer
if ( $value{RASTER_SUB} && $value{RASTER_SUB} ne "none" ) {
    &system_silent("p.select ppm 1>/dev/null");
    &system_silent("p.map.new input=$TMP/$value{PMAP_SUB_SCRIPT} 1> /dev/null");
    &system_errors_logged(
 "cat $ENV{MAPLP} > $TMP/$value{PNM} 2> /dev/null"
    );
# OK to comment out next command if ppmmix not pnmcomb used below
    &system_errors_logged(
"/usr/bin/ppmtopgm < $ENV{MAPLP} > $TMP/$value{PGM} 2> /dev/null"
    );
}

# now create the main raster layer with all vector and site layers
    &system_silent("p.select ppm 1>/dev/null");
    &system_silent("p.map.new input=$TMP/$value{PMAP_SCRIPT} 1> /dev/null");

# only use ppmmix/pnmcomp if a second layer is available to overlay
# ppmmix syntax is ppmmix blend_val underlayer overlayer, with
# a 0.0 giving only the underlayer, 1.0 only overlayer
if ( $value{RASTER_SUB} && $value{RASTER_SUB} ne "none" ) { # 2 layers to overlay
    if ($value{SUB_OPACITY} != 2.0) { # use ppmmix: fixed level opacity
       &system_errors_logged(
"/usr/bin/ppmmix $value{SUB_OPACITY} $TMP/$value{PNM} $ENV{MAPLP} | pnmtopng > $TMP/$value{PNG} 2> /dev/null"
       );
    }
    else { # use pnmcomp: variable opacity
       &system_errors_logged(
"/usr/bin/pnmcomp -alpha=$TMP/$value{PGM} $TMP/$value{PNM} $ENV{MAPLP} | pnmtopng > $TMP/$value{PNG} 2> /dev/null"
       );
    }
}
else { # no RASTER_SUB just the one layer
    &system_errors_logged( "/usr/bin/pnmtopng < $ENV{MAPLP} > $TMP/$value{PNG} 2> /dev/null");
}

} # end create_gif

#-------------------------------------------------------------------------
# NAME: create_display_map
#	Make gif and display it for reclass, combine, buffer operations
# ARGS: $_[0] = any additional text to add to display description
# NOTES: Assumes region settings taken care of by calling function.
#-------------------------------------------------------------------------
sub create_display_map {

    my $addText = $_[0];
    my $baseRaster;

    $value{PMAP_SCRIPT} = "pmap_script.$$";
    $value{PMAP_SUB_SCRIPT} = "pmap_sub_script.$$";

    #$value{DISPLAY_DESC} = "display_desc.$$";
    $value{DISPLAY_DESC} = "gl.dispDesc.$$";
    $value{GIF}          = "grasslinks.$$.gif";
    $value{PNG}          = "grasslinks.$$.png";
    $value{PGM}          = "grasslinks.$$.pgm";
    $value{PNM}          = "grasslinks.$$.pnm";
    $value{GIFSIZE}      = 400;

    open( PMAP_SUB_SCRIPT, ">$TMP/$value{PMAP_SUB_SCRIPT}" )
      || print STDERR
"Unable to open pmap_sub script $TMP/$value{PMAP_SUB_SCRIPT} in sub create_display_map\n";
    print PMAP_SUB_SCRIPT "raster $value{RASTER_SUB}\n end\n";
    close(PMAP_SUB_SCRIPT);

    open( PMAP_SCRIPT, ">$TMP/$value{PMAP_SCRIPT}" )
      || print STDERR
"Unable to open pmap script $TMP/$value{PMAP_SCRIPT} in sub create_display_map\n";
    print PMAP_SCRIPT "raster $value{OUT_RASTER}\n end\n";
    close(PMAP_SCRIPT);

    &create_gif;

    # raster map display description
    if ( $value{RASTER} ) {
        $baseRaster = $RASTERS{ $value{RASTER} }->{descShort};
    }
    else {
        $baseRaster = $RASTERS{ $value{RASTER1} }->{descShort};
        if ( $value{RASTER2} ) {
            $baseRaster .= "</b> and <b>";
            $baseRaster .= $RASTERS{ $value{RASTER2} }->{descShort};
        }
    }

    # Save display description	
    open( DISPLAY_DESC, ">$TMP/$value{DISPLAY_DESC}" );
    print DISPLAY_DESC << "ENDPRINT";
	<FONT SIZE=1 COLOR=GREEN><b>Items Mapped:</b></FONT>\n
        <FONT SIZE=-1>
	<li>New Raster map: <b>$value{OUT_RASTER}</b>\n
	<li>Based on the raster map(s): <b>$baseRaster</b></i>\n
	<li>Within the Region Setting: <b>$REGIONS{ $value{REGION} }->{descShort}</b>\n
	<BR>
	$addText
	</ul></FONT>
ENDPRINT
    close(DISPLAY_DESC);

    # reset the value of $value{RASTER}
    if ( $value{OUT_RASTER} ) {
        $value{RASTER} = $value{OUT_RASTER};
    }
    if ( $value{OUT_RASTER_SUB} ) {
        $value{RASTER_SUB} = $value{OUT_RASTER_SUB};
    }

}

##################################################################
# GRASSLinks Legend subroutines
# ALL of these followng subroutines relate to creating legends for
# raster images; based on the raster's color and category files
# written primarily by James Ganong & require his lookup C program
##################################################################

#-------------------------------------------------------------------------
# NAME: create_legend 
# 	Get the categories information and the color information
# 	to create a legend.
# REQUIRES: a current, valid value for $value{RASTER}
# NOTES: THe call to &find_cats populates the array @catlist and the
#	hash %categories that are used by all the legend subroutines.
#-------------------------------------------------------------------------
sub create_legend {

    if ( !$value{RASTER} || $value{RASTER} eq "none" ) {
        print "<I>No valid raster map selected for legend creation<I><p>\n";
        return;
    }

    # get category information for the RASTER
    my $numcats = &find_cats("$value{RASTER}");

    if ( $numcats < 0 ) {
        print
"<I>Sorry: not enough categories to create an effective legend</I><p>\n";
    }
    elsif ( $numcats > $MAX_LEGEND_CATS ) {
        print
"<I>Sorry: too many categories ($numcats) to create an effective legend</I><p>\n";

    }
    else {

        # create the legend
        open( PMAP_SCRIPT, ">>$TMP/$value{PMAP_SCRIPT}" );
        print PMAP_SCRIPT "legend 2% 98% \n";
        print PMAP_SCRIPT "  height 20\n";    # colorbox height
                                              #print PMAP_SCRIPT "  width 16\n";
            #print PMAP_SCRIPT "  xspace 80\n"; # ignore
            #print PMAP_SCRIPT "  yspace 1\n"; # ignore
        print PMAP_SCRIPT "  textcolor black\n";
        my $textfactor = int( $scalefactor * 0.1 );
        print PMAP_SCRIPT "    textsize $textfactor\n";

        #print PMAP_SCRIPT "  textsize 0.2 \n";
        #print PMAP_SCRIPT "  textwidth 1\n";
        print PMAP_SCRIPT "  background white\n";

        #print PMAP_SCRIPT "  border blue\n";
        print PMAP_SCRIPT "  beginrast\n";

        #print PMAP_SCRIPT "  ramp value vertical\n";
        if ( $value{LEGEND} eq "samp" ) {
            my $catcount = 0;
            foreach $i (@catlist) {
                $catcount++;
                if ( $catcount <= $MAX_LEGEND_CATS ) {
                    print PMAP_SCRIPT " catnum $i $category{$i}\n";
                }
                else {

                    #print "<OPTION VALUE=\"$i\">$category{$i}\n";
                }
            }
        }
        else {
            print ACCESSLOG
"textfactor: $textfactor Cats: $numcats, $category{$numcats}: @catlist\n";
            print PMAP_SCRIPT "    catnum 20 20 m\n";
            print PMAP_SCRIPT "    catnum 200 200 m\n";
            print PMAP_SCRIPT "    catnum 800 800 m\n";
            print PMAP_SCRIPT "    catnum 1600 1600 m\n";
            print PMAP_SCRIPT "    catnum 3200 3200 m\n";
        }
        print PMAP_SCRIPT "    end\n";
        print PMAP_SCRIPT "  end\n";
        print PMAP_SCRIPT "end\n";    # this final end
        close(PMAP_SCRIPT);

        # below is preserved non-working GL 3.1 methods
        #$value{LEGEND_TABLE} = "legend.table.$$";
        #@output= `g.findfile element=colr file=$value{RASTER} mapset=\"\""`;
        #$mapset_for_colortable=$output[1];
        #$mapset_for_colortable =~ s/.*mapset='//s;
        #$mapset_for_colortable =~ s/'.*//s;

        #&get_colors;
        #&make_gifs;
        #&write_legend;
        #&print_file("$TMP/$value{LEGEND_TABLE}");

        # following subroutine needs to be fixed!
        # &write_ranges_legend;
    }

}    # end of create legend

#-------------------------------------------------------------------------
# NAME: get_colors 
#	Gets color values for each category in a GRASS raster file.
# REQUIRES: The C program lookup installed locally. 
# RETURNS: Implicit. Populates 3 hashes - %red{}, %green{}, and %blue{}
# 	that are keyed based on the category value in $catlist[$i].
#-------------------------------------------------------------------------
sub get_colors {

    foreach $i (@catlist) {
        # not working this version of GRASSLINKS
        #( $red{$i}, $green{$i}, $blue{$i} ) =
        #  split ( / /,
            #`$PROG_PATH/lookup $value{RASTER} $mapset_for_colortable $i` );
        #chomp( $blue{$i} );
    }

}

#-------------------------------------------------------------------------
# NAME: make_gifs 
# 	Makes little gif boxes for each legend category.
# REQUIRES: A valid, current @catlist array and associated %red, 
#	%blue, and %green hashes created in &get_colors subroutine.
# CALLS: The subroutine &color2gif to create the actual gif images
#	for each category value.
# RETURNS: Implicit. Populates %gifbox{} hash, keyed by catlist category,
#	with the names of gif files for each category value.
#-------------------------------------------------------------------------
sub make_gifs {

    my $theGif;

    # Saves and restores current region just once instead
    # of each time a legend gif is created
    &system_errors_logged("g.region save=saved_region");
    &system_errors_logged("g.region rast=$LEGEND_RASTER");

    foreach $i (@catlist) {
        $theGif = "$red{$i}.$green{$i}.$blue{$i}.gif";

        # If it doesn't exist, create it
        unless ( -e "$LGIF_FULLPATH/$theGif" ) {
            &color2gif( $red{$i}, $green{$i}, $blue{$i} );
        }
        $gifbox{$i} = "$red{$i}.$green{$i}.$blue{$i}.gif";
    }
    &system_errors_logged("g.region saved_region");

}

#-------------------------------------------------------------------------
# NAME: color2gif 
#	Create little square gifs for legend categories
# ARGS: red, green, and blue values
# REQUIRES: $LEGEND_RASTER and $LEGEND_COLR_TABLE defined, as well
#	as some legend default values. 
#-------------------------------------------------------------------------
sub color2gif {

    my ( $R, $G, $B ) = @_;

    $ENV{HEIGHT} = $LEGEND_GIF_PSIZE;
    $ENV{WIDTH}  = $LEGEND_GIF_PSIZE;

    my $legend_gif = "${R}.${G}.${B}.gif";
    my $theGif     = "$LGIF_FULLPATH/$legend_gif";

    # create the tiny gif to represent legend color
    open( COLOR_TABLE, ">$LEGEND_COLR_TABLE" )
      || print STDERR "cant open colortable in color2gif\n";

    print COLOR_TABLE "#1 first color\n";
    print COLOR_TABLE "0 0 0\n";
    print COLOR_TABLE "$R $G $B\n";
    print COLOR_TABLE "0 0 0\n";
    close COLOR_TABLE;

    # create a pmap script for the legend gif	
    my $SCRIPT = "$TMP/legend.pmap_script.$$";
    open( SCRIPT, ">$SCRIPT" )
      || print STDERR "GL in color2gif can't open script\n";
    print SCRIPT "raster $LEGEND_RASTER\n";
    print SCRIPT "end\n";
    close SCRIPT;

    &system_silent("p.select ppm 1>/dev/null");
    &system_silent("p.map.new input=$SCRIPT 1> /dev/null");
    &system_errors_logged(
"/usr/bin/ppmquant 256 < $ENV{MAPLP} 2> /dev/null  | /usr/bin/ppmtogif -interlace > $LGIF_FULLPATH/$legend_gif 2> /dev/null"
    );

    &flush(STDERR);
}

#-------------------------------------------------------------------------
# NAME: print_legend 
#	Prints out legend in an html table based on values in array
#	@catlist and hashes %category{} and  %gifbox{}.
#-------------------------------------------------------------------------
sub print_legend {

    my $column = 1;

    print '
	<CENTER>
	<TABLE CELLSPACING=4 CELLPADDING=2>
	';

    foreach $i (@catlist) {
        if ( $column == 1 ) {
            print "<TR VALIGN=TOP>\n";
        }

        print "<TD><IMG SRC=\"$LGIF_PATH/$gifbox{$i}\"> $category{$i}</TD>\n";
        if ( $column++ == $MAX_LEGEND_COLS ) {
            print "</TR>\n";
            $column = 1;
        }
    }

    print '
	</TABLE></CENTER>
	<br>
	';

}

#-------------------------------------------------------------------------
# NAME: write_legend 
#	Writes legend in a table format to the file $value{LEGEND_TABLE} 
# REQUIRES: current, valid values for array @catlist, hash %gifbox{},
#	and $MAX_LEGEND_COLS variable.
#-------------------------------------------------------------------------
sub write_legend {

    my $column = 1;

    open( TABLE_TMP, "> $TMP/$value{LEGEND_TABLE}" )
      || print STDERR "GL cant create legend table in write_legend\n";

    print TABLE_TMP "<CENTER>\n";
    print TABLE_TMP "<TABLE CELLSPACING=4 CELLPADDING=2>\n";

    foreach $i (@catlist) {
        if ( $column == 1 ) {
            print TABLE_TMP "<TR VALIGN=TOP>\n";
        }
        print TABLE_TMP
          "<TD> <IMG SRC=\"$LGIF_PATH/$gifbox{$i}\"> $category{$i}</TD>\n";
        if ( $column++ == $MAX_LEGEND_COLS ) {
            print TABLE_TMP "</TR>\n";
            $column = 1;
        }
    }
    print TABLE_TMP "</TABLE>\n";
    print TABLE_TMP "</CENTER><br>\n";
    close(TABLE_TMP);

    &flush(STDERR);

}

#-------------------------------------------------------------------------
# NAME: write_ranges_legend 
#	This is a DRAFT subroutine that is not working and not used!
# 	Should be fixed, important	
#-------------------------------------------------------------------------
sub write_ranges_legend {

    #$colr = `g.filename element=colr file=$value{RASTER} mapset=""`;
    $colr =~ s/file=//;
    $colr =~ s/'//g;
    open( COLOR, "$colr" ) || die "Cannot open color file $colr";
    $maxranges = 24;

    if ( <COLOR> =~ /%/ ) {

        # new style color table, must count ranges
        while (<COLOR>) {

            #$number_of_fields = split;	
            $number_of_fields = 2;
            if ( $number_of_fields != 2 ) {

                # if not two fields it is not a range
                next;
            }

            #$first_field = $_[0];
            #$second_field = $_[1];
            #$number_of_fields = split(/:/,$first_field);
            $number_of_fields = 2;
            $start_range      = $_[0];
            $start_red        = $_[1];
            if ( $number_of_fields == 2 ) {
                $start_green = $start_red;
                $start_blue  = $start_red;
            }
            else {
                $start_green = $_[2];
                $start_blue  = $_[3];
            }

            #$number_of_fields = split(/:/,$second_field);
            $number_of_fields = 2;
            $end_range        = $_[0];
            $end_red          = $_[1];
            if ( $number_of_fields == 2 ) {
                $end_green = $end_red;
                $end_blue  = $end_red;
            }
            else {
                $end_green = $_[2];
                $end_blue  = $_[3];
            }

            if ( $start_range == $end_range ) {

                # we don't want ranges of only one cat
                next;
            }

            if ( $totalranges++ > $maxranges ) {
                print
"too many numeric ranges to create an effective ranges legend\n";
                return;
            }
            else {
                push ( @ranges,
"$start_range:$start_red:$start_green:$start_blue $end_range:$end_red:$end_green:$end_blue"
                );
                push ( @rangesstack,
"$start_range:$start_red:$start_green:$start_blue $end_range:$end_red:$end_green:$end_blue"
                );
                push ( @start_cats, $start_range );
                push ( @end_cats,   $end_range );
                $filename =
"$start_range:$start_red:$start_green:$start_blue-$end_range:$end_red:$end_green:$end_blue";
                push ( @filenames, $filename );
            }
        }    #end of while COLOR?

        if ( $totalranges == 0 ) {
            return;
        }
        $column = 1;
        print "<TABLE>\n";
        foreach $i (@ranges) {
            $args  = pop (@rangesstack);
            $start = pop (@start_cats);
            $end   = pop (@end_cats);
            $file  = pop (@filenames);
            $ENV{'GIF'} = $file;
            &system_errors_logged("$ENV{PROG_PATH}/ramp2gif $args");
            if ( $column == 1 ) { print "<TR VALIGN=TOP>\n"; }
            $thumbnail_url_prefix = "thumb";
            print
"<TD> <IMG SRC=\"$thumbnail_url_prefix/$file\"> $start-$end </TD>\n";

            if ( $column++ == $MAX_LEGEND_COLS ) {
                print "</TR>\n";
                $column = 1;
            }
        }
        print "</TABLE>\n";
    }    #end of if
}    #end of subroutine write_ranges_legend

